
/*     *******************************************************************************
	XVisual GUI Development System for Linux/X-Window
	Copyright (C) 1996 Steve Carrie.

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
        ******************************************************************************* */

/* =============================================
	xvis user interface builder for Athena widgets using
	xv class library.

	CodeGen Segment

	V1.0BETA  - last edited 26/06/96
    ============================================= */

// ---------------------------------------------
// Include file declarations

#include "xvis.h"
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#define GENSOURCESELF		0
#define GENSOURCEFORM		1
#define GENSOURCEMODULE	2

// ---------------------------------------------
// local forward function & data declares

extern func_window_error(Widget aw, XtPointer client, XtPointer call);

void generateMain();
void compileFileList();
void generateMakefile();
void	generateHFiles();
void generateCCFiles();
void generateMainFiles();
int getObjectResources(int obj);
Boolean createResourceValue(String buffer, int thisobj, int i);

void headerLeader(int top, FILE *fp);
void bitmapIncludes(int obj, FILE *fp);
void menuCalls(int top, FILE *fp);
void callScan(int top, FILE *fp);
void callTree(int top, FILE *fp);
void objectCalls(int i, FILE *fp);
void objectScan(int top, FILE *fp, Boolean is_extern, String hdr);
void objectTree(int top, FILE *fp, Boolean is_extern);
void menuTree(int top, FILE *fp, Boolean is_extern);

void CCheader(int top, FILE *fp);
void CCmenuCalls(int top, FILE *fp);
void CCmenu(int top, FILE *fp);
void CCcallScan(int top, FILE *fp);
void CCcallTree(int top, FILE *fp);
void CCobjectCalls(int i, FILE *fp);
void CCobjectTree(int top, FILE *fp);
Boolean CCobjectWrite(int i, FILE *fp, String buffer);

void 	CCwindowObject(int i, FILE *fp, Boolean geninit);

void buildMakeWindow();

char appname[80];
char makefilename[80];
char addlibs[80];
Boolean makefile=FALSE, sharedlink=FALSE;
Boolean makeQuit=FALSE;
Boolean custom=FALSE;
Boolean makecodeonly=FALSE;

char apph[80];
char appcc[80];
int obcount=0;

char *gmenu_build[MAXMENUENTRY];
char *gmenu_names[MAXMENUENTRY];
char *gmenu_subs[MAXMENUENTRY];

int gmenu_ptr[MAXMENUENTRY];
int gmenu_num[MAXMENUENTRY];
int m_num=0;
int makeproc, compstat=0;
Boolean stopMake=FALSE;

// --------------------------------------------
// Callbacks

void func_makechoice(Widget aw, XtPointer client, XtPointer call)
{
}

void func_gen_ok(Widget aw, XtPointer client, XtPointer call)
{
	char *s;

	XtVaGetValues(AppName.ID(), XtNstring, &s, NULL);
	strcpy(appname, s);

	XtVaGetValues(MakeFile.ID(), XtNstring, &s, NULL);
	strcpy(makefilename, s);

	XtVaGetValues(Libs.ID(), XtNstring, &s, NULL);
	strcpy(addlibs, s);

	if (MakeYes.GetState())
		makefile=TRUE;
	else
		makefile=FALSE;

	if (MakeShared.GetState())
		sharedlink=TRUE;
	else
		sharedlink=FALSE;

	xvisGenerate.Close();

	sprintf(apph, "%s.h", appname);
	sprintf(appcc, "%s.cc", appname);
	
	generateMain();
}

void func_gen_cancel(Widget aw, XtPointer client, XtPointer call)
{
	xvisGenerate.Close();
}
 
//---------------------------------------------
// Local Stuff

void commenceCodeGeneration(Boolean codeonly)
{
	makecodeonly=codeonly;
	MakeYes.On();
	if (strlen(project_name)==0) 
		strcpy(appname, "noname");	
	else 
		strcpy(appname, project_name);

	xvisGenerateMenu.EntryResource("Build", XtNsensitive, TRUE);
	xvisGenerateMenu.EntryResource("ReBuild All", XtNsensitive, TRUE);
	
	char *s=strchr(appname, '.');
	if (s!=NULL) *s='\0';

	sprintf(makefilename, "Makefile.%s", appname);	

	AppName.SetResource(XtNstring, (XtArgVal) appname);
	MakeFile.SetResource(XtNstring, (XtArgVal) makefilename);
	xvisGenerate.Open();
}

// --------------------------------------------

void buildCodeGenWindow()
{
	strcpy(addlibs, "");

	xvisGenerate.init("Code Generator", "xvisGenerate", xvisContext);
	xvisGenerate.SetBackground("gray70");
	xvisGenerate.SetWindowResource(XtNgeometry, (XtArgVal)"+200+220");
 	xvisGenerate.SetWindowResource(XtNiconPixmap, 
		xvisGenerate.SetPixmapInfo(xvis_xpm, NULL));

	AppNameLabel.init("Application name", xvisGenerate);
	AppNameLabel.SetBackground("gray70");
	AppNameLabel.SetForeground("blue");
	AppNameLabel.SetDistance(10, 20);
	AppNameLabel.ChangeFont("*times-*-r-*-140*"); 
	AppNameLabel.SetBorderWidth(0);
	AppNameLabel.SetFixedPosition(TRUE);

	AppName.init("AppName", xvisGenerate, XawtextEdit, 20, 150);
	AppName.SetDistance(130, 20);
	XtTranslations appname_trans=XtParseTranslationTable(
		"<Key>Return: select-all() \n\
		!Meta<Key> i: no-op()");
	AppName.OverrideTranslations(appname_trans);
	AppName.SetFixedPosition(TRUE);

	LinkChoiceLabel.init("Link Mode", xvisGenerate);
	LinkChoiceLabel.SetBackground("gray70");
	LinkChoiceLabel.SetForeground("blue");
	LinkChoiceLabel.SetDistance(10, 60);
	LinkChoiceLabel.ChangeFont("*times-*-r-*-140*"); 
	LinkChoiceLabel.SetBorderWidth(0);
	LinkChoiceLabel.SetFixedPosition(TRUE);

	MakeShared.init("Shared", xvisGenerate, func_makechoice, "Shared");
	MakeShared.SetDistance(150, 60);
	MakeShared.ChangeFont("*times-*-r-*-140*"); 
	MakeShared.SetShadowWidth(4);
	MakeShared.SetSize(60, 30);
	MakeShared.SetFixedPosition(TRUE);
	MakeShared.SetBeNiceToColormap(FALSE);

	MakeStatic.init("Static", xvisGenerate, func_makechoice, "Static");
	MakeStatic.SetDistance(210, 60);
	MakeStatic.ChangeFont("*times-*-r-*-140*"); 
	MakeStatic.SetShadowWidth(4);
	MakeStatic.SetSize(60, 30);
	MakeStatic.SetFixedPosition(TRUE);
	MakeStatic.ChangeGroup(MakeShared);
	MakeShared.State(TRUE);
	MakeStatic.SetBeNiceToColormap(FALSE);

	MakeShared.SetSensitive(FALSE);
	MakeStatic.SetSensitive(FALSE);

	MakeChoiceLabel.init("Generate Makefile", xvisGenerate);
	MakeChoiceLabel.SetBackground("gray70");
	MakeChoiceLabel.SetForeground("blue");
	MakeChoiceLabel.SetDistance(10, 100);
	MakeChoiceLabel.ChangeFont("*times-*-r-*-140*"); 
	MakeChoiceLabel.SetBorderWidth(0);
	MakeChoiceLabel.SetFixedPosition(TRUE);

	MakeYes.init("YES", xvisGenerate, func_makechoice, "YES");
	MakeYes.SetDistance(150, 100);
	MakeYes.ChangeFont("*times-*-r-*-140*"); 
	MakeYes.SetShadowWidth(4);
	MakeYes.SetSize(50, 30);
	MakeYes.SetFixedPosition(TRUE);
	MakeYes.SetBeNiceToColormap(FALSE);

	MakeNo.init("NO", xvisGenerate, func_makechoice, "NO");
	MakeNo.SetDistance(200, 100);
	MakeNo.ChangeFont("*times-*-r-*-140*"); 
	MakeNo.SetShadowWidth(4);
	MakeNo.SetSize(50, 30);
	MakeNo.SetFixedPosition(TRUE);
	MakeNo.ChangeGroup(MakeYes);
	MakeYes.State(TRUE);
	MakeNo.SetBeNiceToColormap(FALSE);

	MakeFileLabel.init("Makefile name", xvisGenerate);
	MakeFileLabel.SetBackground("gray70");
	MakeFileLabel.SetForeground("blue");
	MakeFileLabel.SetDistance(10, 140);
	MakeFileLabel.ChangeFont("*times-*-r-*-140*"); 
	MakeFileLabel.SetBorderWidth(0);
	MakeFileLabel.SetFixedPosition(TRUE);

	MakeFile.init("MakeFile", xvisGenerate, XawtextEdit, 20, 150);
	MakeFile.SetDistance(130, 140);
	XtTranslations makefile_trans=XtParseTranslationTable(
		"<Key>Return: select-all() \n\
		!Meta<Key> i: no-op()");
	MakeFile.OverrideTranslations(makefile_trans);
	MakeFile.SetFixedPosition(TRUE);
	MakeFile.SetBeNiceToColormap(FALSE);

	LibLabel.init("Additional Libs", xvisGenerate);
	LibLabel.SetBackground("gray70");
	LibLabel.SetForeground("blue");
	LibLabel.SetDistance(10, 180);
	LibLabel.ChangeFont("*times-*-r-*-140*"); 
	LibLabel.SetBorderWidth(0);
	LibLabel.SetFixedPosition(TRUE);

	Libs.init("Libs", xvisGenerate, XawtextEdit, 20, 150);
	Libs.SetDistance(130, 180);
	XtTranslations lib_trans=XtParseTranslationTable(
		"<Key>Return: select-all() \n\
		!Meta<Key> i: no-op()");
	Libs.OverrideTranslations(lib_trans);
	Libs.SetFixedPosition(TRUE);
	Libs.SetBeNiceToColormap(FALSE);

	genOK.init("Generate", xvisGenerate, func_gen_ok);
	genOK.SetDistance(45, 250);
	genOK.ChangeFont("*times-*-r-*-140*"); 
	genOK.SetWidth(75);
	genOK.SetShadowWidth(4);
	genOK.SetBackground("gray85");
	genOK.SetFixedPosition(TRUE);
	genOK.SetBeNiceToColormap(FALSE);

	genCancel.init("Cancel", xvisGenerate, func_gen_cancel);
	genCancel.SetDistance(140, 250);
	genCancel.ChangeFont("*times-*-r-*-140*"); 
	genCancel.SetWidth(75);
	genCancel.SetShadowWidth(4);
	genCancel.SetBackground("gray85");
	genCancel.SetFixedPosition(TRUE);
	genCancel.SetBeNiceToColormap(FALSE);

}

// --------------------------------------------
// --------------------------------------------

void generateMain()
{
	compileFileList();
	if (numccfiles > 0) {
		if (makefile) generateMakefile();
		generateHFiles();
		generateCCFiles();
		generateMainFiles();
		if (!makecodeonly) buildProgram(TRUE);
	}
}

// --------------------------------------------

void compileFileList()
{
	custom=FALSE;
	numccfiles=0;
	numhfiles=0;
	char tmp[80];

	for (int i=0; i<MAXOBJECTS; i++)
		if (objects[i].object_type==XVWindow) {
			sprintf(h_files[numhfiles].filename, "%s.h", objects[i].name);
			strcpy(h_files[numhfiles].stubname, objects[i].name);
			h_files[numhfiles].source=i;
			h_files[numhfiles++].sourcetype=GENSOURCEFORM;

			sprintf(cc_files[numccfiles].filename, "%s.cc", objects[i].name);
			strcpy(cc_files[numccfiles].stubname, objects[i].name);
			cc_files[numccfiles].source=i;
			cc_files[numccfiles++].sourcetype=GENSOURCEFORM;
		}
	
	for (int i=0; i<MAXEDITWINDOWS; i++)
		if (module[i].in_use) {
			strcpy(tmp, module[i].name);
			if (strstr(tmp, ".h")!=NULL) {
				strcpy(h_files[numhfiles].filename, tmp);
				char *s=strchr(tmp, '.');
				if (s!=NULL) *s='\0';
				strcpy(h_files[numhfiles].stubname, tmp);
				h_files[numhfiles].source=i;
				h_files[numhfiles++].sourcetype=GENSOURCEMODULE;
			}
			else {
				strcpy(cc_files[numccfiles].filename, tmp);
				char *s=strchr(tmp, '.');
				if (s!=NULL) *s='\0';
				strcpy(cc_files[numccfiles].stubname, tmp);
				if ((strcmp(tmp, "Custom")==0) || (strcmp(tmp, "custom")==0))
					custom=TRUE;
				cc_files[numccfiles].source=i;
				cc_files[numccfiles++].sourcetype=GENSOURCEMODULE;
			}
		}
	
}

// ---------------------------------------------

void generateMakefile()
{
	FILE *fp;
	fp=fopen(makefilename, "w");
	if (fp != NULL) {
		fprintf(fp, "# Makefile generated by xvis\n\n");
		if (sharedlink)
			fprintf(fp, "CC=g++\nCFLAGS= -O\nINCLUDES= -I.\n");
		else
			fprintf(fp, "CC=g++\nCFLAGS= -static -O\nINCLUDES= -I.\n");
		fprintf(fp, "BIN26LINKOPTS=-u gettxt -L/usr/X11/lib -L/usr/i486-linux/lib\n");
		fprintf(fp, "ADDLIBS = %s \n", addlibs);
		fprintf(fp, "XLIBS = -lxv -lXaw3d -lXpm -lXmu -lXt -lSM -lICE -lXext -lX11 \n\n");
		fprintf(fp, "OBJFILES=\t%s.o ", appname);
		for (int i=0; i<numccfiles; i++)
			fprintf(fp, "%s.o ", cc_files[i].stubname);
		fprintf(fp, "\n\n");
		fprintf(fp, ".cc.o:\n\t$(CC) -c $(CFLAGS) $(INCLUDES) $*.cc\n\n");
		fprintf(fp, "all::%s\n\n", appname);
		if (sharedlink)
			fprintf(fp, "%s:	$(OBJFILES)\n\t$(CC) $(BIN26LINKOPTS) -o $@ $(OBJFILES) $(XLIBS) $(ADDLIBS)\n\n", 
				appname);
		else
			fprintf(fp, "%s:	$(OBJFILES)\n\t$(CC) $(BIN26LINKOPTS) -static -o $@ $(OBJFILES) $(XLIBS)\n\n", 
				appname);
		fprintf(fp, "clean:\n\trm -f *.o %s\n\n", appname);
		fprintf(fp, "remake:: clean all\n\n");
		fclose(fp);
	}
}

// ----------------------------------------------

void generateHFiles()
{
	FILE *fp;

	for (int i=0; i<numhfiles; i++) 
		if (h_files[i].sourcetype==GENSOURCEFORM) {
			fp=fopen(h_files[i].filename, "w");
			if (fp != NULL) {
				headerLeader(i, fp);
				bitmapIncludes(h_files[i].source, fp);
				if (objects[h_files[i].source].hasmenu) menuCalls(h_files[i].source, fp);
				callScan(h_files[i].source, fp);
				objectScan(h_files[i].source, fp, TRUE, 
					"Export objects from this form to other modules.");
				fprintf(fp, "\n#endif // _%s_h_\n\n", h_files[i].stubname);
				fclose(fp);
			}
		}
}

// ----------------------------------------------
// ----------------------------------------------

void headerLeader(int top, FILE *fp)
{
	fprintf(fp, "/* ************************************************** \n");
	fprintf(fp, "                    Header file for  %s\n", h_files[top].stubname);
	fprintf(fp, "                    Generated by xvis\n");
	fprintf(fp, "**************************************************** */\n\n");  
	fprintf(fp, "#ifndef _%s_h_\n", h_files[top].stubname);
	fprintf(fp, "#define _%s_h_\n\n", h_files[top].stubname);
	fprintf(fp, "/* Export form initialisation header */\n\n");
	fprintf(fp, "extern void _init_%s();\n", objects[h_files[top].source].name);
}

// ----------------------------------------------

void bitmapIncludes(int obj, FILE *fp)
{
	fprintf(fp, "\n/* *******************************************************\n");
	fprintf(fp, "      Include any bitmap files required bit internal objects.\n");
	fprintf(fp,  "********************************************************* */\n\n");

	for (int i=0; i<MAXOBJECTS; i++) 
		if (objects[i].parent_window == obj)
			switch (objects[i].object_type) {
				case XVWindow:
					if (objects[i].has_bitmap) {
						fprintf(fp, "#ifndef _pixmap_%s_\n", objects[i].obj_bitref);
						fprintf(fp, "\t#define _pixmap_%s_\n", objects[i].obj_bitref);
						fprintf(fp, "\t#include \"%s\"\n", objects[i].obj_icon);
						fprintf(fp, "#endif //  _pixmap_%s_\n", objects[i].obj_bitref);
					}
					break;
				default:
					if (objects[i].has_bitmap) {
						fprintf(fp, "#ifndef _pixmap_%s_\n", objects[i].obj_bitref);
						fprintf(fp, "\t#define _pixmap_%s_\n", objects[i].obj_bitref);
						fprintf(fp, "\t#include \"%s\"\n", objects[i].obj_path);
						fprintf(fp, "#endif //  _pixmap_%s_\n", objects[i].obj_bitref);
					}
					if (objects[i].has_leftbitmap) {
						fprintf(fp, "#ifndef _pixmap_%s_\n", objects[i].obj_leftbitref);
						fprintf(fp, "\t#define _pixmap_%s_\n", objects[i].obj_leftbitref);
						fprintf(fp, "\t#include \"%s\"\n", objects[i].leftbitmap);
						fprintf(fp, "#endif //  _pixmap_%s_\n", objects[i].obj_leftbitref);
					}
					break;
			}
}

// ----------------------------------------------

void menuCalls(int top, FILE *fp)
{
	m_num=-1;
	if (objects[top].totalmenu > 0) {
		fprintf(fp, "\n/* *******************************************************\n");
		fprintf(fp, "   menu function protoypes for this form, extern'ed for other modules\n");
		fprintf(fp,  "********************************************************* */\n\n");

		for (int i=0; i<objects[top].totalmenu; i++) {
			if (objects[top].menuinfo[i].type == XVMENUMAIN) {
				m_num++;
				gmenu_build[m_num]=objects[top].menuinfo[i].caption;
				gmenu_names[m_num]=objects[top].menuinfo[i].name;
				gmenu_num[m_num]=0;
				gmenu_ptr[m_num]=i+1;
			}
			if (objects[top].menuinfo[i].type == XVMENUSEP) 
				gmenu_num[m_num]++;
			
			if (objects[top].menuinfo[i].type == XVMENUSUB) {
				gmenu_num[m_num]++;
				fprintf(fp, 
					"extern void %s (Widget aw, XtPointer clientdata, XtPointer calldata);\n", 
					objects[top].menuinfo[i].callback);
			}
		}
		m_num++;
	}
}

// ----------------------------------------------

void callScan(int top, FILE *fp)
{
	fprintf(fp, "\n/* *******************************************************\n");
	fprintf(fp, "    call protoypes for this form, extern'ed for other modules\n");
	fprintf(fp,  "********************************************************* */\n\n");

	objectCalls(top, fp);
	callTree(top, fp);
}

// ----------------------------------------------

void callTree(int top, FILE *fp)
{
	objectType ob;

	for (int i=0; i<MAXOBJECTS; i++) {
		ob=objects[i].object_type;
		if ((i!=top) &&(ob !=XVNone))
			if (objects[i].parent_index == top ) {
				objectCalls(i, fp);
				if ((ob== XVWindow) || (ob==XVBox) || (ob==XVForm))
					callTree(i, fp);
			}	
	}	
}

// ----------------------------------------------

void copyLine(String source, String dest)
{
	char c;
	for (int i=0; i<strlen(source); i++) {
		c=source[i];
		if ((c=='\n') || (c=='\0')) {
			dest[i]='\0';
			break;			
		}
		else
			dest[i]=c;
	}
}

// ----------------------------------------------

void objectCalls(int i, FILE *fp)
{
	char tmp[180];

	for (int j=0; j<NUMEVENTS; j++) 
		if ((objects[i].events & eventinfo[j].eventbits) != 0) {
			sprintf(tmp, 
				"extern void %s_%s (Widget aw, XtPointer client_data, %s *event, Boolean *cont);", 
				objects[i].name, eventinfo[j].eventName, 
				eventinfo[j].eventStruct);
			fprintf(fp, "%s\n", tmp);				
		}

	callType *calltype=object_call_types[objects[i].object_type];
	char **callname=object_call_names[objects[i].object_type];
	int num=object_call_number[objects[i].object_type];
	
	if (num>0) {
		for (int j=0; j<num; j++) {
			switch(calltype[j]) {
				case XVCallback:
					sprintf(tmp, 
						"extern void %s_%s (Widget aw, XtPointer clientdata, XtPointer calldata);", 
						objects[i].name, callname[j]);
					break;
				case XVTimeout:
					sprintf(tmp, 
						"extern void %s_%s (XtPointer calldata, XtIntervalId *id);", 
						objects[i].name, callname[j]);
					break;
			}
			fprintf(fp, "%s\n", tmp);							
		}
	}
}

// ----------------------------------------------

void objectScan(int top, FILE *fp, Boolean is_extern, String hdr)
{
	fprintf(fp, "\n/* *******************************************************\n");
	fprintf(fp, "        %s\n", hdr);
	fprintf(fp,  "********************************************************* */\n\n");

	if (is_extern)
		fprintf(fp, "extern xvWindow %s;\n", objects[top].name);
	else
		fprintf(fp, "xvWindow %s;\n", objects[top].name);

	objectTree(top, fp, is_extern);
	if ((objects[top].hasmenu) && (m_num>0)) menuTree(top, fp, is_extern);
}

// ----------------------------------------------

void objectTree(int top, FILE *fp, Boolean is_extern)
{
	objectType ob;

	for (int i=0; i<MAXOBJECTS; i++) {
		ob=objects[i].object_type;
		if ((i!=top) &&(ob !=XVNone))
			if (objects[i].parent_index == top ) {
				if (is_extern)
					fprintf(fp, "\textern %s %s;\n", className[ob], objects[i].name);
				else
					fprintf(fp, "\t%s %s;\n", className[ob], objects[i].name);
				if ((ob== XVWindow) || (ob==XVBox) || (ob==XVForm))
					objectTree(i, fp, is_extern);
			}	
	}	

}

// ----------------------------------------------

void menuTree(int top, FILE *fp, Boolean is_extern)
{
	if (is_extern) 
		fprintf(fp, "\nextern xvMenuBar %s_menubar;\n", objects[top].name);	
	else
		fprintf(fp, "\nxvMenuBar %s_menubar;\n", objects[top].name);	
	for (int i=0; i<m_num; i++)
		if (is_extern)
			fprintf(fp, "\textern xvPulldownMenu %s_%s;\n", 
				objects[top].name, gmenu_names[i]);	
		else
			fprintf(fp, "\txvPulldownMenu %s_%s;\n", 
				objects[top].name, gmenu_names[i]);	
	
	if (is_extern) 
		fprintf(fp, "\n\textern char *%s_menubar_data[];\n", objects[top].name);
	else	{
		fprintf(fp, "\n\tchar *%s_menubar_data[] = {\t", objects[top].name);
		for (int i=0; i<m_num; i++) {
			if (i!=0) fprintf(fp, ", ");
			fprintf(fp, "\"%s\"", gmenu_build[i]);
		}
		fprintf(fp, "};\n");
	}
	int r=0;	
	for (int i=0; i<m_num; i++) 
		if (is_extern)	
			fprintf(fp, "\textern char *%s_%s_data[];\n", 
				objects[top].name, gmenu_names[i]);
		else {
			fprintf(fp, "\tchar *%s_%s_data[] = {\t", 
				objects[top].name, gmenu_names[i]);
			r=gmenu_ptr[i];	
			for (int j=0; j<gmenu_num[i]; j++) {
				if (j!=0) fprintf(fp, ", ");
				fprintf(fp, "\"%s\"", objects[top].menuinfo[r+j].caption);
			}
			fprintf(fp, "};\n");
		}	
	for (int i=0; i<m_num; i++) 
		if (is_extern)	
			fprintf(fp, "\textern XtCallbackProc %s_%s_proc[];\n", 
				objects[top].name, gmenu_names[i]);
		else {
			fprintf(fp, "\tXtCallbackProc %s_%s_proc[] = {\t", 
				objects[top].name, gmenu_names[i]);
			r=gmenu_ptr[i];	
			for (int j=0; j<gmenu_num[i]; j++) {
				if (j!=0) fprintf(fp, ", ");
				fprintf(fp, "%s", objects[top].menuinfo[r+j].callback);
			}
			fprintf(fp, "};\n");
		}	
}

// ----------------------------------------------

void generateCCFiles()
{
	FILE *fp;

	for (int i=0; i<numccfiles; i++) 
		if (cc_files[i].sourcetype==GENSOURCEFORM) {
			fp=fopen(cc_files[i].filename, "w");
			if (fp != NULL) {
				CCheader(i, fp);
				CCcallScan(cc_files[i].source, fp);
				if (objects[cc_files[i].source].hasmenu) CCmenuCalls(cc_files[i].source, fp);
				
				objectScan(cc_files[i].source, fp, FALSE, 
					"Objects for this form.");

				CCwindowObject(i, fp, TRUE);

				if (objects[cc_files[i].source].hasmenu) CCmenu(cc_files[i].source, fp);
				
				CCobjectTree(cc_files[i].source, fp);

				CCwindowObject(i, fp, FALSE);

				fprintf(fp, "}\n\n// End of %s \n\n", cc_files[i].filename);
				fclose(fp);
			}
		}
}

// -----------------------------------------------

void CCheader(int top, FILE *fp)
{
	fprintf(fp, "/* ************************************************** \n");
	fprintf(fp, "                    Code file for  %s\n", cc_files[top].stubname);
	fprintf(fp, "                    Generated by xvis\n");
	fprintf(fp, "**************************************************** */\n\n");  
	fprintf(fp, "/* Import application header */\n");
	fprintf(fp, "#include %c%s%c\n\n", 34, apph, 34);
}

// -----------------------------------------------

void CCmenu(int top, FILE *fp)
{
	Pixel temppix;

	fprintf(fp, "\n/* *******************************************************\n");
	fprintf(fp, "   				Form control Menus\n");
	fprintf(fp,  "********************************************************* */\n\n");
	
	fprintf(fp, "\t%s_menubar.init(\"%s_menubar\", %s, %d, %s_menubar_data);\n",
		objects[top].name, objects[top].name, 
		objects[top].name, objects[top].nummenu, objects[top].name);
	fprintf(fp, "\t%s_menubar.SetDistance(0, 0);\n", objects[top].name);
	fprintf(fp, "\t%s_menubar.SetDefaultDistance(0);\n", objects[top].name);
	fprintf(fp, "\t%s_menubar.SetFixedPosition(TRUE);\n", objects[top].name);
	fprintf(fp, "\t%s_menubar.SetHeight(35);\n", objects[top].name);
	fprintf(fp, "\t%s_menubar.SetAllResource(XtNshadowWidth, 0);\n", objects[top].name);
	fprintf(fp, "\t%s_menubar.ChangeFont(\"%s\");\n", 
		objects[top].name, objects[top].obj_menu_font);
	fprintf(fp, "\t%s_menubar.SetAllResource(XtNresizable, TRUE);\n", objects[top].name);
	fprintf(fp, "\t%s_menubar.SetResizable(TRUE);\n", objects[top].name);
	fprintf(fp, "\t%s_menubar.SetBorderWidth(0);\n", objects[top].name);
			
	for (int i=0; i<m_num; i++) {
		fprintf(fp, "\t\t%s_%s.init(\"%s\", %d, %s_%s_data, %s_%s_proc, %s_menubar.GetButton(\"%s\"));\n",
			objects[top].name, gmenu_names[i],
			gmenu_build[i],
			objects[top].numpulldown[i],
			objects[top].name, gmenu_names[i],
			objects[top].name, gmenu_names[i],
			objects[top].name,
			gmenu_build[i]);
		fprintf(fp, "\t\t%s_%s.ChangeFont(\"%s\");\n", 
			objects[top].name, gmenu_names[i],
			objects[top].obj_menu_font);
	}
	fflush(fp);
}

// -----------------------------------------------

void CCmenuCalls(int top, FILE *fp)
{

	char *tmp=new char [MAXEDITBUFFERSIZE];

	if (objects[top].totalmenu > 0) {
		fprintf(fp, "\n/* *******************************************************\n");
		fprintf(fp, "   menu functions for this form, extern'ed for other modules\n");
		fprintf(fp,  "********************************************************* */\n\n");

		for (int i=0; i<objects[top].totalmenu; i++) 
			if (objects[top].menuinfo[i].type == XVMENUSUB) { 
				if (objects[top].menuinfo[i].codebuffer == NULL)
					sprintf(tmp, 
						"void %s (Widget aw, XtPointer clientdata, XtPointer calldata)\n{\n\n}\n", 
						objects[top].menuinfo[i].callback);
				else
					strcpy(tmp, objects[top].menuinfo[i].codebuffer);
			
			fprintf(fp, "%s\n", tmp);
			}
		
	}
	fflush(fp);
	delete tmp;
}

// -----------------------------------------------

void CCcallScan(int top, FILE *fp)
{
	fprintf(fp, "\n/* *******************************************************\n");
	fprintf(fp, "    calls for this form\n");
	fprintf(fp,  "********************************************************* */\n\n");
	
	CCobjectCalls(top, fp);
	CCcallTree(top, fp);
}

// -----------------------------------------------

void CCcallTree(int top, FILE *fp)
{
	objectType ob;

	for (int i=0; i<MAXOBJECTS; i++) {
		ob=objects[i].object_type;
		if ((i!=top) && (ob != XVNone))
			if (objects[i].parent_index==top) {
				CCobjectCalls(i, fp);
				if ((ob==XVWindow) || (ob==XVBox) || (ob==XVForm))
					CCcallTree(i, fp);
			}
	}
}

// -----------------------------------------------

void CCobjectCalls(int i, FILE *fp)
{
	char *tmp=new char[MAXEDITBUFFERSIZE];

	for (int j=0; j<NUMEVENTS; j++) 
		if ((objects[i].events & eventinfo[j].eventbits) != 0) {
			if (objects[i].eventcode[j]==NULL)
				sprintf(tmp, 
					"void %s_%s (Widget aw, XtPointer client_data, %s *event, Boolean *cont)\n{\n\n}\n", 
					objects[i].name, eventinfo[j].eventName, 
					eventinfo[j].eventStruct);
			else
				strcpy(tmp, objects[i].eventcode[j]);
				
			fprintf(fp, "%s\n", tmp);				
		}

	callType *calltype=object_call_types[objects[i].object_type];
	char **callname=object_call_names[objects[i].object_type];
	int num=object_call_number[objects[i].object_type];
	
	if (num>0) {
		for (int j=0; j<num; j++) {
			switch(calltype[j]) {
				case XVCallback:
					if (objects[i].callbackcode[j]==NULL)
						sprintf(tmp, 
							"void %s_%s (Widget aw, XtPointer clientdata, XtPointer calldata)\n{\n\n}\n", 
							objects[i].name, callname[j]);
					else
						strcpy(tmp, objects[i].callbackcode[j]);

					break;
				case XVTimeout:
					if (objects[i].callbackcode[j]==NULL)
						sprintf(tmp, 
							"void %s_%s (XtPointer calldata, XtIntervalId *id)\n{\n\t%s.Reset();\n}\n", 
							objects[i].name, callname[j], objects[i].name);
					else
						strcpy(tmp, objects[i].callbackcode[j]);

					break;
			}
			fprintf(fp, "%s\n", tmp);							
		}
	}
	fflush(fp);
	delete tmp;
}

// -----------------------------------------------

void 	CCwindowObject(int i, FILE *fp, Boolean geninit)
{
	char *buffer=new char[2048];

	int obj=cc_files[i].source;
	if (geninit) {
		fprintf(fp, "\n/* *******************************************************\n");
		fprintf(fp, "    Init Function for %\n", cc_files[i].stubname);
		fprintf(fp,  "********************************************************* */\n\n");

		fprintf(fp, "void _init_%s()\n{\n\n", cc_files[i].stubname);

		fprintf(fp, "//\tInit window %s\n\n", objects[obj].name);
		fprintf(fp, "\t%s.init(\"%s\", \"xvWindow\", appContext);\n\n", 
			objects[obj].name, 
			objects[obj].name);
			
		int n=getObjectResources(obj);
	
		for (int j=0; j<n; j++) {
			if (createResourceValue(buffer, obj, j))
				fprintf(fp, "%s\n", buffer);
		}
	}
	else {
		fprintf(fp, "\n/* *******************************************************\n");
		fprintf(fp, "    Set Size Functions for %s\n", cc_files[i].stubname);
		fprintf(fp,  "********************************************************* */\n\n");
		int n=getObjectResources(obj);
	
		for (int j=n-5; j<n-2; j++) {
			if (createResourceValue(buffer, obj, j))
				fprintf(fp, "%s\n", buffer);
		}
	}

	if (geninit) 
		for (int j=0; j<NUMEVENTS; j++)
			if ((objects[obj].events & eventinfo[j].eventbits) != 0) { 
				sprintf(buffer, 
				"\t%s.AddEvent(%s, (XtEventHandler) %s_%s, (XtPointer) \"%s\");", 
				objects[obj].name, eventinfo[j].eventMask, 
				objects[obj].name, eventinfo[j].eventName, 
				objects[obj].name);
				fprintf(fp, "%s\n", buffer);				
			}

	fflush(fp);
	delete buffer;
}

// -----------------------------------------------

void CCobjectTree(int top, FILE *fp)
{
	char *buffer=new char[2048];

	objectType ob;
	int n;

	for (int i=0; i<MAXOBJECTS; i++) {
		ob=objects[i].object_type;
		if ((i!=top) && (ob != XVNone)) {
			if (objects[i].parent_index==top) {
				if (CCobjectWrite(i, fp, buffer)) {
					fprintf(fp, "\n//\tInit %s %s\n\n", objectName[objects[i].object_type], 
						objects[i].name);
					fprintf(fp, "%s\n", buffer);
					if (ob != XVTimer) 
						if ((ob==XVFileList) || (ob == XVDirList) || (ob==XVList))
							fprintf(fp, "\t%s.xvView::SetFixedPosition(TRUE);\n", objects[i].name);
						else
							fprintf(fp, "\t%s.SetFixedPosition(TRUE);\n", objects[i].name);
					if (uses3D[ob])
						fprintf(fp, "\t%s.SetBeNiceToColormap(FALSE);\n", objects[i].name);
					fflush(fp);
					n=getObjectResources(i);
					for (int j=0; j<n; j++) {
						if (createResourceValue(buffer, i, j)) {
							fprintf(fp, "%s\n", buffer);
							fflush(fp);	
						}	
					}
					for (int j=0; j<NUMEVENTS; j++)
						if ((objects[i].events & eventinfo[j].eventbits) != 0) { 
							sprintf(buffer, 
							"\t%s.AddEvent(%s, (XtEventHandler) %s_%s, (XtPointer) \"%s\");", 
							objects[i].name, eventinfo[j].eventMask, 
							objects[i].name, eventinfo[j].eventName, 
							objects[i].name);
							fprintf(fp, "%s\n", buffer);				
						}

				}
				if ((ob==XVWindow) || (ob==XVBox) || (ob==XVForm))
					CCobjectTree(i, fp);
			}
		}
	}		
	fflush(fp);
	delete buffer;
}

// -----------------------------------------------

Boolean CCobjectWrite(int i, FILE *fp, String buffer)
{
	buffer[0]='\0';

	char parent[30];
	char obname[30];

	Boolean valid=TRUE;
	objectType otype=objects[i].object_type;
	strcpy(parent, objects[objects[i].parent_index].name);
	strcpy(obname, objects[i].name);

	switch (otype) {
		case XVTimer:
			sprintf(buffer, "\t%s.init(\"%s\", %s, %dL, %s_timeout, FALSE);\n",
				obname, obname, parent, objects[i].object_time, obname);			
			break;
		case XVList:
			sprintf(buffer, "\t%s.init(\"%s\", %s, %s_select, 0, empty_list, 1, FALSE);",
				obname, obname, parent, obname, FALSE);
			break;
		case XVButton:
			sprintf(buffer, "\t%s.init(\"%s\", %s, %s_click, \"%s\");", 
				obname, obname, parent, obname, obname);
			break;
		case XVRepeater:
			sprintf(buffer, "\t%s.init(\"%s\", %s, %s_click, \"%s\", %s_start, %s_stop);", 
				obname, obname, parent, obname, obname, obname, obname);
			break;
		case XVBitmap:
			sprintf(buffer, "\t%s.init(\"%s\", %s, TRUE);",
				obname, obname, parent);
			break;
		case XVLabel:
			sprintf(buffer, "\t%s.init(\"%s\", %s);", 
				obname, obname, parent);
			break;
		case XVDropDown:
			sprintf(buffer, "\t%s.init(\"%s\", %s, 1, empty_list, %s_click, %s_select, 5, 0);",
				obname, obname, parent, obname, obname);
			break;
		case XVEditField:
			sprintf(buffer, "\t%s.init(\"%s\", %s, %s, 100, 100, XawAsciiString, NULL, TRUE, %s_change, \"%s\");",
				obname, obname, parent, editmodedata[objects[i].edit_mode], obname, obname);
			break;
		case XVBox:
			sprintf(buffer, "\t%s.init(\"%s\", %s);", obname, obname, parent);
			break;
		case XVToggle:
			sprintf(buffer, "\t%s.init(\"%s\", %s, %s_click, \"%s\");",
				obname, obname, parent, obname, obname);
			break;
		case XVForm:
			sprintf(buffer, "\t%s.init(\"%s\", %s, TRUE, TRUE);",
				obname, obname, parent);
			break;
		case XVVScroll:
			sprintf(buffer, "\t%s.init(\"%s\", %s, XtorientVertical, 100, %s_scroll, %s_thumb, %s_jump);",
				obname, obname, parent, obname, obname, obname);
			break;
		case XVHScroll:
			sprintf(buffer, "\t%s.init(\"%s\", %s, XtorientHorizontal, 100, %s_scroll, %s_thumb, %s_jump);",
				obname, obname, parent, obname, obname, obname);
			break;
		case XVPanner:
			sprintf(buffer, "\t%s.init(\"%s\", %s, %s_move, 100, 100, 50, 50);",
				obname, obname, parent, obname);
			break;
		case XVFileList:
			sprintf(buffer, "\t%s.init(\"%s\", %s, \"%s\", XVFILELIST, FALSE, %s_select, TRUE);",
				obname, obname, parent, objects[i].obj_path, obname);
			break;
		case XVDirList:
			sprintf(buffer, "\t%s.init(\"%s\", %s, \"%s\", XVDIRLIST, FALSE, %s_select, TRUE);",
				obname, obname, parent, objects[i].obj_path, obname);
			break;
		default:
			valid=FALSE;
			break;
	}
	return(valid);
}

// -----------------------------------------------
// -----------------------------------------------

void generateMainFiles()
{
	FILE *fp;
	
	fp=fopen(apph, "w");
	if (fp != NULL) {
		fprintf(fp, "/* ********************************************************\n");
		fprintf(fp, "                    Main header file for application %s\n", appname);
		fprintf(fp, "                    Generated by xvis\n");
		fprintf(fp, "********************************************************* */\n\n");
		fprintf(fp, "#ifndef _%s_h_\n", appname);
		fprintf(fp, "#define _%s_h_\n\n", appname);
		fprintf(fp, "/* Include xv library header file */\n\n");
		fprintf(fp, "#include <X11/xv/xv.h>\n");
		fprintf(fp, "#include <stdio.h>\n");
		fprintf(fp, "#include <unistd.h>\n\n");
		fprintf(fp, "/* Include all header files for this application */\n");
		for (int i=0; i<numhfiles; i++)
			fprintf(fp, "#include %c%s%c\n", 34, h_files[i].filename, 34);
		fprintf(fp, "\n");
		fprintf(fp, "/* Global context */\n");
		fprintf(fp, "extern xvContext appContext;\n");
		fprintf(fp, "extern Boolean _appExit;\n");
		fprintf(fp, "extern char *empty_list[];\n\n");
		
		fprintf(fp, "extern Display *appDisplay;\n");
		fprintf(fp, "extern Screen *appScreen;\n");
		fprintf(fp, "extern int screenNumber, screenWidth, screenHeight, screenDepth;\n");
		fprintf(fp, "extern unsigned long screenMask;\n");
		fprintf(fp, "extern Pixel blackPixel, whitePixel;\n");
		fprintf(fp, "extern char appCwd[];\n\n");

		fprintf(fp, "extern void appExit(int retcode=0);\n\n");
		fprintf(fp, "extern void _app_dummy_callback(Widget aw, XtPointer client, XtPointer call);\n");
		fprintf(fp, "extern void func_dummy(Widget aw, XtPointer client, XtPointer call);\n");
		fprintf(fp, "extern void _app_dummy_event(Widget aw, XtPointer client, register XEvent *event, Boolean *cont);\n");
		fprintf(fp, "extern void _app_dummy_timeout(XtPointer cdata, XtIntervalId *id);\n\n");
		if (custom) {
			fprintf(fp, "extern void appCustomInit(int argc, char **argv);\n\n");
			fprintf(fp, "extern void appCustomClose();\n\n");
		}
		fprintf(fp, "/* End of %s */\n\n", apph);
		fprintf(fp, "#endif // _%s_h_\n\n", appname);
		fclose(fp);
	}	

	fp=fopen(appcc, "w");
	if (fp != NULL) {
		fprintf(fp, "/* ********************************************************\n");
		fprintf(fp, "                    Main app file for application %s\n", appname);
		fprintf(fp, "                    Generated by xvis\n");
		fprintf(fp, "********************************************************* */\n\n");

		fprintf(fp, "/* Include app header file */\n");
		fprintf(fp, "#include %c%s%c\n\n", 34, apph, 34);

		fprintf(fp, "/* Global vars for this section */\n");
		fprintf(fp, "xvContext appContext;\n");
		fprintf(fp, "Boolean _appExit=FALSE;\n");
		fprintf(fp, "char *empty_list[]= { \"empty_list\" };\n\n");

		fprintf(fp, "Display *appDisplay;\n");
		fprintf(fp, "Screen *appScreen;\n");
		fprintf(fp, "int screenNumber, screenWidth, screenHeight, screenDepth;\n");
		fprintf(fp, "unsigned long screenMask;\n");
		fprintf(fp, "Pixel blackPixel, whitePixel;\n");
		fprintf(fp, "char appCwd[180];\n\n");

		fprintf(fp, "// Dummy functions\n");
		fprintf(fp, "void _app_dummy_callback(Widget aw, XtPointer client, XtPointer call) { }\n");
		fprintf(fp, "void func_dummy(Widget aw, XtPointer client, XtPointer call) { }\n");
		fprintf(fp, "void _app_dummy_event(Widget aw, XtPointer client, register XEvent *event, Boolean *cont) { }\n");
		fprintf(fp, "void _app_dummy_timeout(XtPointer cdata, XtIntervalId *id) { }\n\n");

		fprintf(fp, "// Init functions\n");
		fprintf(fp, "void _appInit(int argc, char **argv)\n{\n");

		fprintf(fp, "\tappDisplay=XOpenDisplay(NULL);\n");
		fprintf(fp, "\tappScreen=DefaultScreenOfDisplay(appDisplay);\n");
		fprintf(fp, "\tscreenNumber=DefaultScreen(appDisplay);\n");
		fprintf(fp, "\tscreenWidth=WidthOfScreen(appScreen);\n");
		fprintf(fp, "\tscreenHeight=HeightOfScreen(appScreen);\n");
		fprintf(fp, "\tscreenDepth=DefaultDepthOfScreen(appScreen);\n");
		fprintf(fp, "\twhitePixel=WhitePixel(appDisplay, screenNumber);\n");
		fprintf(fp, "\tblackPixel=BlackPixel(appDisplay, screenNumber);\n");
		fprintf(fp, "\tscreenMask=AllPlanes;\n");
		fprintf(fp, "\tXCloseDisplay(appDisplay);\n");
		fprintf(fp, "\tgetcwd(appCwd, 180);\n\n");

		fprintf(fp, "\tappContext.init(%c%s%c, argc, argv);\n\n", 34, appname, 34);
		for (int i=0; i<numhfiles; i++)
			if (h_files[i].sourcetype==GENSOURCEFORM)
				fprintf(fp, "\t_init_%s();\n", objects[h_files[i].source].name);	
		if (custom) {
			fprintf(fp, "\n//\tCall user's custom init routine\n");
			fprintf(fp, "\tappCustomInit(argc, argv);\n\n");
		}
		fprintf(fp, "}\n\n");

		fprintf(fp, "// appExit function\n");
		fprintf(fp, "void appExit(int retcode)\n{\n");
		fprintf(fp, "\t_appExit=TRUE;\n");
		for (int i=0; i<MAXOBJECTS; i++)
			if (objects[i].object_type == XVWindow)
				fprintf(fp, "\t%s.Close();\n", objects[i].name);
		if (custom) {
			fprintf(fp, "\n//\tCall user's custom close routine\n");
			fprintf(fp, "\tappCustomClose();\n\n");
		}
		fprintf(fp, "\n\texit(retcode);\n");
		fprintf(fp, "}\n\n");				

		fprintf(fp, "// Main\n");
		fprintf(fp, "main (int argc, char **argv)\n");
		fprintf(fp, "{\n\n//\tCall init functions\n\n");
		fprintf(fp, "\t_appInit(argc, argv);\n\n");
		fprintf(fp, "\n//\tOpen startup windows... \n\n");
		for (int i=0; i<MAXOBJECTS; i++)
			if ((objects[i].object_type == XVWindow) && (objects[i].open_at_startup))
				fprintf(fp, "\t%s.Open();\n", objects[i].name);
		fprintf(fp, "\n//\tMain loop\n\n");
		fprintf(fp, "\tXtAppMainLoop(appContext.GetContext());\n\n");
		fprintf(fp, "}\n\n");
		fprintf(fp, "// End of %s\n\n", appcc);
		fclose(fp);
	}
}
 
// -----------------------------------------------
// -----------------------------------------------

int getObjectResources(int obj)
{
	objectType otype=objects[obj].object_type;
	int numres=object_resource_number[otype];
	char* *types=object_resource_types[otype];
	resourceType *datatypes=object_data_types[otype];

	for (int i=0; i<numres; i++) {
		resource_values[i].r_restype=types[i];
		resource_values[i].r_type=datatypes[i];
	}
	recoverResources(numres, obj);		
	return(numres);
}

// -----------------------------------------------

Boolean createResourceValue(String buffer, int thisobj, int i)
{
	for (int z=0; z<strlen(buffer); z++)
		buffer[z]='\0';

	char obname[40];
	strcpy(obname, objects[thisobj].name);

	int otype=objects[thisobj].object_type;
	int ref;
	Boolean valid=TRUE;

	switch(resource_values[i].r_type) {
		case XVPixel:
			if (strcmp(resource_values[i].r_restype, XtNbackground)==0) {
				if (strcmp(objects[thisobj].background_colour, "(default)")!=0) 
					sprintf(buffer, "\t%s.SetBackground(\"%s\");", obname, 
						objects[thisobj].background_colour);
				else
					valid=FALSE;
			}
			if (strcmp(resource_values[i].r_restype, XtNforeground)==0) {
				if (strcmp(objects[thisobj].foreground_colour, "(default)")!=0) 
					sprintf(buffer, "\t%s.SetForeground(\"%s\");", obname, 
						objects[thisobj].foreground_colour);
				else
					valid=FALSE;
			}
			if (strcmp(resource_values[i].r_restype, XtNborderColor)==0) {
				if (strcmp(objects[thisobj].border_colour, "(default)")!=0) 
					sprintf(buffer, "\t%s.SetBorderColor(\"%s\");", obname, 
						objects[thisobj].border_colour);
				else
					valid=FALSE;
			}
			if (strcmp(resource_values[i].r_restype, XtNtopShadowPixel)==0) {
				if (strcmp(objects[thisobj].topshadow_colour, "(default)")!=0) 
					sprintf(buffer, "\t%s.SetTopShadowPixel(\"%s\");", obname, 
						objects[thisobj].topshadow_colour);
				else
					valid=FALSE;
			}
			if (strcmp(resource_values[i].r_restype, XtNbottomShadowPixel)==0) {
				if (strcmp(objects[thisobj].bottomshadow_colour, "(default)")!=0) 
					sprintf(buffer, "\t%s.SetBottomShadowPixel(\"%s\");", obname, 
						objects[thisobj].bottomshadow_colour);
				else
					valid=FALSE;
			}
			break;
		case XVDimension:
			switch(otype) {
				case XVList:
				case XVFileList:
				case XVDirList:
					sprintf(buffer, 
						"\t%s.xvView::SetResource(\"%s\", (XtArgVal) %d);", 
						obname, 
						resource_values[i].r_restype, 
						resource_values[i].r_dimension);
					break;
				case XVDropDown:
					if (strcmp(resource_values[i].r_restype, XtNwidth)==0) {
						sprintf(buffer, 
						"\t%s.ChangeWidth(%d);", 
						obname, resource_values[i].r_dimension);
						break;
					}
				default: 
					sprintf(buffer,
						"\t%s.SetResource(\"%s\", (XtArgVal) %d);",  
						obname, 
						resource_values[i].r_restype, 
						resource_values[i].r_dimension);
					break;
			}
			break;
		case XVPosition:
			switch(otype) {
				case XVList:
				case XVFileList:
				case XVDirList:
					sprintf(buffer, 
						"\t%s.xvView::SetResource(\"%s\", (XtArgVal) %d);", 
						obname, 
						resource_values[i].r_restype, 
						resource_values[i].r_position);
					break;
				default:
					sprintf(buffer,
						"\t%s.SetResource(\"%s\", (XtArgVal) %d);",  
						obname, 
						resource_values[i].r_restype, 
						resource_values[i].r_position);
					break;
			}
			break;
		case XVInt:
			sprintf(buffer, 	"\t%s.SetResource(\"%s\", (XtArgVal) %d);",  
				obname,
				resource_values[i].r_restype, 
				resource_values[i].r_int);
			break;
		case XVString:
			sprintf(buffer, 	"\t%s.SetResource(\"%s\", (XtArgVal) \"%s\");",  
				obname,
				resource_values[i].r_restype, 
				resource_values[i].r_string);
			break;
		case XVCursor:
			if (strcmp(objects[thisobj].obj_cursor, "(None)")!=0) 
				sprintf(buffer, 	"\t%s.SetCursorName(\"%s\");",   
					objects[thisobj].obj_cursor);
			else
				valid=FALSE;
			break;
		case XVBoolean:
			if (resource_values[i].r_boolean == TRUE)					
				strcpy(resource_values[i].r_value, "TRUE");
			else
				strcpy(resource_values[i].r_value, "FALSE");
			sprintf(buffer, 	"\t%s.SetResource(\"%s\", (XtArgVal) \"%s\");",   
				obname,
				resource_values[i].r_restype, 
				resource_values[i].r_value);
			break;
		case XVLocBoolean:
			if (strcmp(resource_values[i].r_restype, xvNfixed)==0) 
				if (objects[thisobj].fixed_size == TRUE) 
					sprintf(buffer, "\t%s.FixBorder(%d, %d);", 
					obname, objects[thisobj].winwidth, objects[thisobj].winheight);
				else
					valid=FALSE;
							
			if (strcmp(resource_values[i].r_restype, xvNiconic)==0) 
				if (objects[thisobj].iconic == TRUE) 
					sprintf(buffer, "\t%s.SetWindowResource(XtNiconic, TRUE);",
						obname);	
				else
					valid=FALSE;

			if (strcmp(resource_values[i].r_restype, xvNopenAtStart)==0) 
				valid==FALSE;

			if (strcmp(resource_values[i].r_restype, xvNenabled)==0) 
				if (objects[thisobj].enabled == FALSE) 
					sprintf(buffer, "\t%s.SetSensitive(FALSE);",
						obname);	
				else
					valid=FALSE;

			if (strcmp(resource_values[i].r_restype, xvNvisible)==0) 
				if (objects[thisobj].visible == FALSE) 
					sprintf(buffer, "\t%s.SetMappedWhenManaged(FALSE);",
						obname);	
				else
					valid=FALSE;

			if (strcmp(resource_values[i].r_restype, xvNctrl)==0) 
				if (objects[thisobj].ctrl == TRUE) 
					sprintf(buffer, "\t%s.DisplayChars(TRUE);",
						obname);	
				else
					valid=FALSE;

			if (strcmp(resource_values[i].r_restype, xvNecho)==0) 
				if (objects[thisobj].echo == FALSE) 
					sprintf(buffer, "\t%s.Echo(FALSE);",
						obname);	
				else
					valid=FALSE;

			if (strcmp(resource_values[i].r_restype, xvNautowrap)==0) 
				if (objects[thisobj].autowrap == TRUE) 
					sprintf(buffer, "\t%s.AutoWrap(TRUE);",
						obname);	
				else
					valid=FALSE;

			break;
		case XVWinName:
			sprintf(buffer, 
				"\t%s.SetWindowResource(XtNtitle, (XtArgVal) \"%s\");\n\t%s.SetWindowResource(XtNiconName, (XtArgVal) \"%s\");",
						obname, resource_values[i].r_string,
						obname, resource_values[i].r_string);	
			break;
		case XVWinSize:
			if (strcmp(resource_values[i].r_restype, XtNheight)==0) 
				sprintf(buffer, 
					"\t%s.SetWindowResource(XtNgeometry, (XtArgVal) \"%dx%d\");", 
					obname, objects[thisobj].winwidth, objects[thisobj].winheight);
			else 		
				valid=FALSE;
			break;
		case XVWinPositionX:
			sprintf(buffer,
				"\t%s.SetWindowResource(XtNgeometry, (XtArgVal) \"+%d+%d\");", 
				obname, objects[thisobj].winxpos, objects[thisobj].winypos);
			break;
		case XVFont:
			if (strcmp(objects[thisobj].obj_font, "*fixed-*-r-*-100-*-70*")!=0)
				sprintf(buffer, 
					"\t%s.ChangeFont(\"%s\");",
					obname, objects[thisobj].obj_font);
			else
				valid=FALSE;
			break;
		case XVPath:
			switch (otype) {
				case XVFileList:
					sprintf(buffer, "\t%s.SetNewPath(\"%s\", XVFILELIST, FALSE);",
						obname, objects[thisobj].obj_path);
					break;
				case XVDirList:
					sprintf(buffer, "\t%s.SetNewPath(\"%s\", XVDIRLIST, FALSE);",
						obname, objects[thisobj].obj_path);
					break;
				default:
					valid=FALSE;
					buffer[0]='\0';
					break;
			}
			break;
		case XVBitmapFile:
			switch (otype) {
				case XVWindow:
					if (strcmp(objects[thisobj].obj_icon, "(None)")!=0) {
						if (strstr(objects[thisobj].obj_icon, ".xpm")==NULL)
							sprintf(buffer, 
								"\t%s.SetWindowResource(XtNiconPixmap, %s.SetBitmapInfo(%s_bits, %s_width, %s_height, NULL));",
									obname, obname, 
									objects[thisobj].obj_bitref,
									objects[thisobj].obj_bitref,
									objects[thisobj].obj_bitref);
						else
							sprintf(buffer, 
								"\t%s.SetWindowResource(XtNiconPixmap, %s.SetPixmapInfo(%s_xpm, NULL));",
									obname, obname, 
									objects[thisobj].obj_bitref);
					}
					else
						valid=FALSE;
					break;
				default:
					if (strcmp(resource_values[i].r_restype, xvNbitmapFile)==0)	{				
						if (strcmp(objects[thisobj].obj_path, "(None)")!=0) {
							if (strstr(objects[thisobj].obj_path, ".xpm")==NULL)
								sprintf(buffer, 
									"\t%s.SetBitmapInfo(%s_bits, %s_width, %s_height, XtNbitmap);",
									obname, 
									objects[thisobj].obj_bitref,
									objects[thisobj].obj_bitref,
									objects[thisobj].obj_bitref);
							else
								sprintf(buffer, "\t%s.SetPixmapInfo(%s_xpm, XtNbitmap);",
									obname,
									objects[thisobj].obj_bitref);
						}
						else
							valid=FALSE;
					}
					else {
						if (strcmp(objects[thisobj].leftbitmap, "(None)")!=0) {
							if (strstr(objects[thisobj].leftbitmap, ".xpm")==NULL)
								sprintf(buffer, 
									"\t%s.SetBitmapInfo(%s_bits, %s_width, %s_height, XtNleftBitmap);",
									obname, 
									objects[thisobj].obj_leftbitref,
									objects[thisobj].obj_leftbitref,
									objects[thisobj].obj_leftbitref);
							else
								sprintf(buffer, "\t%s.SetPixmapInfo(%s_xpm, XtNleftBitmap);",
									obname,
									objects[thisobj].obj_leftbitref);
						}
						else
							valid=FALSE;
					}
					break;
			}
			break;
		case XVObject:
			if (strcmp(resource_values[i].r_value, "(None)")!=0) 
				sprintf(buffer, "\t%s.ChangeGroup(%s);",
					obname, objects[objects[thisobj].object_ref3].name);
			break;
		case XVhScroll:
			sprintf(buffer, "\t%s.HorizontalScrollBar(%s);",
				obname, scrolldata[objects[thisobj].hscroll_mode]);
			break;
		case XVvScroll:
			sprintf(buffer, "\t%s.VerticalScrollBar(%s);",
				obname, scrolldata[objects[thisobj].vscroll_mode]);
			break;
		case XVWrap:
			sprintf(buffer, "\t%s.Wrap(%s);",
				obname, wrapdata[objects[thisobj].wrap_mode]);
			break;
		case XVResize:
			sprintf(buffer, "\t%s.ResizeMode(%s);",
				obname, resizedata[objects[thisobj].resize_mode]);
			break;
		case XVShapeStyle:
			if (objects[thisobj].shape != 0) 
				sprintf(buffer, "\t%s.SetResource(XtNshapeStyle, (XtArgVal) %s);", 
					obname, buttonshapedata[objects[thisobj].shape]);
			break;
		case XVJustify:
			if (objects[thisobj].justify != 0) 
				sprintf(buffer, "\t%s.SetJustify(%s);", 
					obname, justifydata[objects[thisobj].justify]);
			break;
		default:
			valid=FALSE;
			break;
	}
	return(valid);
}

// -------------------------------------------------

void make_quit(Widget aw, XtPointer client, XtPointer call)
{
	alert.Close();
	xvisMake.Close();
	strcpy(errorbuffer, makebuffer);
	if (! ErrorsVisible) 
		func_window_error(NULL, NULL, NULL);
	delete makebuffer;
}

// -------------------------------------------------

void getMakeInput(XtPointer client, int *fid, XtInputId *id)
{
	char buf [2048];
	int nbytes=0;
	int i, j;

	if ((nbytes=read(*fid, buf, 80)) != -1) 
		if (nbytes>0) {
			buf[nbytes]='\0';
			if (strlen(makebuffer)+strlen(buf) < MAXEDITBUFFERSIZE) {
				strcat(makebuffer, buf);
				maketext.SetResource(XtNstring, (XtArgVal) makebuffer);
			}
		}
	else {
		j=waitpid(-1, &compstat, WNOHANG | WUNTRACED);
		if (j!=0) {
			makequit.SetSensitive(FALSE);
			stopMake=TRUE;
		}
	}
}

// -------------------------------------------------

void buildProgram(Boolean buildall)
{
	makebuffer=new char[MAXEDITBUFFERSIZE];
	FILE *rp;

	int j=0;
	makeQuit=FALSE;
	stopMake=FALSE;

	XtInputId iid;

	char makefilecmd[180];
	char errname[50];
	char msg[40];
	char longmsg[80];

	if (ErrorsVisible) 
		func_window_error(NULL, NULL, NULL);

	buildMakeWindow();
	xvisMake.Open();

	sprintf(makebuffer, "Make Operation for %s\n\n", appname);
	
	sprintf(errname, "errs.%s", appname);
	if (buildall)
		sprintf(makefilecmd, "make -f %s remake 2>> %s 1>> %s", makefilename, errname, errname);
	else	
		sprintf(makefilecmd, "make -f %s 2>> %s 1>> %s", makefilename, errname, errname);

	remove(errname);
	rp=fopen(errname, "w");
	fclose(rp);
	maketext.SetResource(XtNstring, (XtArgVal) makebuffer);
	makeproc=fork();
	if (makeproc==0) int y=execl("/bin/sh", "sh", "-c", makefilecmd, NULL);

	rp=fopen(errname, "r");

	if (rp!=NULL) 
		iid=XtAppAddInput(xvisContext.GetContext(), fileno(rp), (XtPointer) XtInputReadMask, 
			getMakeInput, NULL);
 
	while ((!makeQuit) && (! stopMake)) {
		j=waitpid(-1, &compstat, WNOHANG | WUNTRACED);
		if (j!=0) break;
		xvisContext.ServiceEvent();	
	}

	if (makeQuit) kill(makeproc, SIGKILL);

	int retstat=-1;
	if (WIFEXITED(compstat) !=0 ) retstat=WEXITSTATUS(compstat);

	if (rp!=NULL) {
		fflush(rp);
		fclose(rp);	
		XtRemoveInput(iid);
	}
	
	if (makeQuit) {
		strcpy(msg, "Make Operation cancelled");
		strcpy(longmsg, "Make Operation cancelled.\nProgram compilation cancelled\nby user.");
		strcat(makebuffer, "Make Operation ABORTED.\n\n");
		xvisGenerateMenu.EntryResource("Run", XtNsensitive, FALSE);
	}
	else {
		strcpy(msg, "Make Operation complete");
		switch(retstat) {
			case -1:
				strcpy(longmsg, "Make Operation completed.\nCompile status could not be determined.\nCheck error window for details.");
				xvisGenerateMenu.EntryResource("Run", XtNsensitive, FALSE);
				break;
			case 0:
				strcpy(longmsg, "Make Operation completed.\nProgram appears to have compiled correctly\n Check error window for details.");
				xvisGenerateMenu.EntryResource("Run", XtNsensitive, TRUE);
				break;
			default:
				strcpy(longmsg, "Make Operation completed.\n There were errors during compilation.\n Check error window for details.");
				xvisGenerateMenu.EntryResource("Run", XtNsensitive, FALSE);
				break;
		}
		strcat(makebuffer, "Make Operation COMPLETE.\n\n");
	}

	alert.init(msg, xvisControl, longmsg,
	XVMSGSIMPLEWARN, "yellow",
	make_quit, make_quit);
	
	alert.Open();
}

 // -------------------------------------------------

void makequit_call(Widget aw, XtPointer client, XtPointer call)
{
	makeQuit=TRUE;
}

// -------------------------------------------------

void buildMakeWindow()
{
	xvisMake.init("Make", "makewin", xvisContext);
	xvisMake.SetBackground("gray85");

	maketext.init("maketext", xvisMake, XawtextRead,  400, 600);
	maketext.ChangeFont("*times-*-r-*-140*");
	maketext.SetBackground("LemonChiffon");
	maketext.Wrap(XawtextWrapLine);
	
	makequit.init("Close", xvisMake, makequit_call);
	makequit.ChangeFont("*times-*-r-*-140*");
	makequit.SetSize(80, 30);
	makequit.SetShadowWidth(4);
	makequit.SetRelVertical(maketext);
	makequit.SetHorizDistance(260);
	makequit.SetBeNiceToColormap(FALSE);
}

// --------------------------------------------------

void runProgram()
{
	char runfilecmd[180];
	if (strlen(appname)>0) {
		sprintf(runfilecmd, "%s", appname);
		makeproc=fork();
		if (makeproc==0) int y=execl("/bin/sh", "sh", "-c", runfilecmd, NULL);
	}
}

// ---------------------------------------------------
