
/*     *******************************************************************************
	XVisual GUI Development System for Linux/X-Window
	Copyright (C) 1996 Steve Carrie.

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
        ******************************************************************************* */

/* =============================================
	xvis user interface builder for Athena widgets using
	xv class library.

	xvisFile segment

	V1.0  - last edited 01/07/96
    ============================================= */

// ---------------------------------------------
// Local Include file declarations

#include "xvis.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>

// ---------------------------------------------
// Local function declarations

int fileExist(String name);
int SaveCurrentForm(String name);
int SaveObject(FILE *fp, int obj);
int SaveObjectTree(FILE *fp, int obj);
int LoadObject(FILE *fp, int obj, int index);
int LoadProject();
int LoadCurrentForm(String name);
int LoadObjectTree(FILE *fp, int toplevel, int *index, int count);
void CreateLoadedObject(int thisobject, objectType object);
int ReadString(FILE *fp, String datastring);
void SetLoadedResources(int obj);
void RelinkReferences(int index);
void ShowInfoWindow(String infostring);
void HideInfoWindow();

void func_choice_yes(Widget aw, XtPointer client, XtPointer call);
void func_choice_no(Widget aw, XtPointer client, XtPointer call);

extern Boolean xlocked;
extern int loadCurrentModule(String name);
extern void winexpose(Widget aw, XtPointer client, register XExposeEvent *event, Boolean *cont);
extern int maptype;

// ---------------------------------------------
// Local Data

int filemode;
char currentfilename[180];
xvPopupShell info;
xvLabel infolabel;

// ---------------------------------------------
// Local callbacks

void func_choice_yes(Widget aw, XtPointer client, XtPointer call)
{
	alert.Close();
	SaveCurrentForm(currentfilename);
	xlocked=FALSE;
}

void func_choice_no(Widget aw, XtPointer client, XtPointer call)
{
	alert.Close();
	xlocked=FALSE;
}

// ---------------------------------------------
// Info Window control.

void ShowInfoWindow(String infostring)
{
	info.init("info", xvisControl);
	infolabel.init("infolabel", info);
	infolabel.ChangeFont("*times-*-r-*-140*");
	infolabel.SetLabel(infostring);
	info.Open();
}

void HideInfoWindow()
{
	info.Close();
}
 
// ---------------------------------------------
// File Action

void execFileAction(int fileAction, String name)
{
	strcpy(currentfilename, name);
	switch(fileAction) {
		case xvFORMSAVEAS:
			if (fileExist(name)!=-1) {
				alert.init("File Exists", xvisControl, 
					"The specified file exists.\n Do you want to Overwrite?",
					XVMSGSIMPLECHOICE, "green",
					func_choice_yes, func_choice_no);
				alert.Open();
			}
			else {
				SaveCurrentForm(name);
				xlocked=FALSE;
			}
			break;
		case xvFORMSAVE:
			SaveCurrentForm(name);
			xlocked=FALSE;
			break;
		case xvFORMLOAD:
			if (fileExist(name)==-1) {
				alert.init("File Does Not Exist", xvisControl, 
					"The specified file does not exist.",
					XVMSGSIMPLEFATAL, "red",
					func_choice_no, func_choice_no);
				alert.Open();
			}
			else {
				LoadCurrentForm(name);
				xlocked=FALSE;
			}
			break;
		case xvPROJECTLOAD:
			if (fileExist(project_path)==-1) {
				alert.init("Project Does Not Exist", xvisControl, 
					"The specified project does not exist.",
					XVMSGSIMPLEFATAL, "red",
					func_choice_no, func_choice_no);
				alert.Open();
			}
			else 
				ShowInfoWindow("Loading Project...");
				LoadProject();
				if (strlen(project_name)>0)
					xvisFileMenu.EntryResource("Save Project", XtNsensitive, TRUE);
				HideInfoWindow();
			break;
		default:
			break;
	}
}

// ---------------------------------------------
// Does file exist?

int fileExist(String name)
{
	int filestat;
	struct stat s;
	filestat=stat(name, &s);
	filemode=s.st_mode;
	return(filestat);
}	

// ---------------------------------------------
// Save current form as 'name'

int SaveCurrentForm(String name)
{
	FILE *fp;

	int toplevel=objects[CurrentObject].parent_window;
	int err=0;

	fp=fopen(name, "w");

	if (fp!=NULL) {
		fprintf(fp, "Xvisual Version 1.0.1\n");
		err=SaveObject(fp, toplevel);
		err=SaveObjectTree(fp, toplevel);
		fprintf(fp, "--End-Of-File--\n");
		fclose(fp);
		objects[toplevel].changed=FALSE;
	}	
}

// --------------------------------------
// Save Object Tree

int SaveObjectTree(FILE *fp, int toplevel)
{
	objectType ob;
	int err=0;

	for (int i=0; i<MAXOBJECTS; i++) {
		ob=objects[i].object_type;
		if ((i!=toplevel) && (ob !=XVNone))
			if (objects[i].parent_index == toplevel) {
				err=SaveObject(fp, i);
				if ((ob==XVBox) || (ob=XVForm)) SaveObjectTree(fp, i);
			}
	}				
}

// --------------------------------------
// Save an object

int SaveObject(FILE *fp, int obj)
{
	objectType otype=objects[obj].object_type;
	int numres=object_resource_number[otype];
	char* *types=object_resource_types[otype];
	resourceType *datatypes=object_data_types[otype];

	for (int i=0; i<numres; i++) {
		resource_values[i].r_restype=types[i];
		resource_values[i].r_type=datatypes[i];
	}
	recoverResources(numres, obj);		

	fprintf(fp, "-Object Start-\n");
	fprintf(fp, "%s\n", objects[obj].name);
	fprintf(fp, "%s\n", objects[obj].obj_cursor);
	fprintf(fp, "%s\n", objects[obj].obj_font);
	fprintf(fp, "%s\n", objects[obj].obj_path);
	fprintf(fp, "%s\n", objects[obj].obj_icon);
	fprintf(fp, "%s\n", objects[obj].leftbitmap);
	fprintf(fp, "%s\n", objects[obj].obj_menu_font);
	fprintf(fp, "%s\n", objects[obj].background_colour);
	fprintf(fp, "%s\n", objects[obj].foreground_colour);
	fprintf(fp, "%s\n", objects[obj].border_colour);
	fprintf(fp, "%s\n", objects[obj].topshadow_colour);
	fprintf(fp, "%s\n", objects[obj].bottomshadow_colour);
	fprintf(fp, "%d %d\n", objects[obj].object_type, objects[obj].parent_type);
	if (objects[obj].parent_index == -1)
		fprintf(fp, "(None)\n");
	else
		fprintf(fp, "%s\n", objects[objects[obj].parent_index].name);

	fprintf(fp, "%s\n", objects[objects[obj].parent_window].name);
	fprintf(fp, "%ld\n", objects[obj].events);
	fprintf(fp, "%d\n", objects[obj].num_children);
	fprintf(fp, "%d %d\n", objects[obj].winxpos, objects[obj].winypos);
	fprintf(fp, "%d %d\n", objects[obj].winwidth, objects[obj].winheight);
	fprintf(fp, "%d %d\n", objects[obj].fixed_size, objects[obj].iconic);
	fprintf(fp, "%d\n", objects[obj].open_at_startup);
	fprintf(fp, "%d %d\n", objects[obj].enabled, objects[obj].visible);

	fprintf(fp, "%d %d %d\n", objects[obj].ctrl, objects[obj].echo, objects[obj].autowrap);
	fprintf(fp, "%d %d %d %d %d %d %d\n", objects[obj].vscroll_mode, objects[obj].hscroll_mode, 
					objects[obj].wrap_mode, objects[obj].resize_mode,
					objects[obj].edit_mode, objects[obj].shape,
					objects[obj].justify);

	if (objects[obj].object_ref1 != -1)
		fprintf( fp, "%s\n", objects[objects[obj].object_ref1].name);
	else
		fprintf(fp, "None\n");

	if (objects[obj].object_ref2 != -1)
		fprintf( fp, "%s\n", objects[objects[obj].object_ref2].name);
	else
		fprintf(fp, "None\n");

	if (objects[obj].object_ref3 != -1)
		fprintf( fp, "%s\n", objects[objects[obj].object_ref3].name);
	else
		fprintf(fp, "None\n");

	fprintf(fp, "%d\n", objects[obj].object_time);
	fprintf(fp, "%d\n", numres);

	for (int i=0; i<numres; i++) 
		fprintf(fp, "%s\n", resource_values[i].r_value);

	fprintf(fp, "%d %d %d\n", objects[obj].hasmenu, objects[obj].totalmenu, 
		objects[obj].nummenu);

	fprintf(fp, "%s\n", objects[obj].menu_background);
	fprintf(fp, "%s\n", objects[obj].menu_foreground);

	if (objects[obj].hasmenu) {

		if (objects[obj].nummenu > 0) 
			for (int i=0; i<objects[obj].nummenu; i++)
				fprintf(fp, "%d\n", objects[obj].numpulldown[i]);

		if (objects[obj].totalmenu > 0) 
			for (int i=0; i<objects[obj].totalmenu; i++) {
				fprintf(fp, "%d\n", objects[obj].menuinfo[i].type);
				fprintf(fp, "%d\n", objects[obj].menuinfo[i].listpoint);
				fprintf(fp, "%s\n", objects[obj].menuinfo[i].caption);
				fprintf(fp, "%s\n", objects[obj].menuinfo[i].name);
				fprintf(fp, "%s\n", objects[obj].menuinfo[i].callback);
				if (objects[obj].menuinfo[i].codebuffer == NULL)
					fprintf(fp, "%d\n", 0);
				else {
					fprintf(fp, "%d\n", strlen(objects[obj].menuinfo[i].codebuffer));
					fprintf(fp, "%s", objects[obj].menuinfo[i].codebuffer);			
				}
			}
	}

	for (int i=0; i<NUMEVENTS; i++) 
		if (objects[obj].eventcode[i] != NULL) {
			fprintf(fp, "%d\n", strlen(objects[obj].eventcode[i]));
			fprintf(fp, "%s", objects[obj].eventcode[i]);
		}
		else
			fprintf(fp, "%d\n", 0);

	for (int i=0; i<3; i++) 
		if (objects[obj].callbackcode[i] != NULL) {
			fprintf(fp, "%d\n", strlen(objects[obj].callbackcode[i]));
			fprintf(fp, "%s", objects[obj].callbackcode[i]);
		}
		else
			fprintf(fp, "%d\n", 0);
	
	fprintf(fp, "-Object End-\n");

}

// ---------------------------------------------
// Load current form from 'name'

int LoadCurrentForm(String name)
{
	FILE *fp;
	int err=0, pe;
	char verstr[80];
	maptype=0;
	int level=0, thisobject, formobject;
	Boolean vercheck=TRUE;

	fp=fopen(name, "r");

	if (fp!=NULL) {
		if (nobject<MAXOBJECTS) {
			thisobject=firstAvailableSlot();
			formobject=thisobject;
			ReadString(fp, verstr);
			vercheck=(strcmp(verstr, "Xvisual Version 1.0.3-BETA")==0);
			err=LoadObject(fp, thisobject, level);
			CurrentObject=-1;
			CreateLoadedObject(thisobject, XVWindow);
			SetLoadedResources(thisobject);
			objects[thisobject].changed=FALSE;

			if (objects[thisobject].hasmenu) constructVisualMenu(thisobject, 
				objects[thisobject].totalmenu);
			level++;
			err=LoadObjectTree(fp, thisobject, &level, -1);
			pe=findFirstProjectEntry();
			project_list[pe]=thisobject;
			RelinkReferences(level);
		}
		fclose(fp);
		if (nobject!=0) {
			controlEventsData(TRUE);
			xvisObjectMenu.EntryResource("Remap Object...", XtNsensitive, TRUE);
			xvisObjectMenu.EntryResource("New/Edit Menu", XtNsensitive, TRUE);
			xvisFileMenu.EntryResource("Save Project As...", XtNsensitive, TRUE);
			xvisFileMenu.EntryResource("Save Form As...", XtNsensitive, TRUE);
			xvisFileMenu.EntryResource("Save Project", XtNsensitive, TRUE);
			xvisFileMenu.EntryResource("Save Form", XtNsensitive, TRUE);
			xvisSave.SetSensitive(TRUE);
			xvisEditMenu.SetSensitive(TRUE);
		}
		SetCounterValue();
		updateProjectListWindow();
	}	
}

// --------------------------------------
// Load Object Tree

int LoadObjectTree(FILE *fp, int toplevel, int *index, int count)
{
	int thisobject, err=0;
	int cnum, vcount=0;
	objectType otype;

	CurrentObject=toplevel;
	CurrentClass=objects[CurrentObject].pcore;

	while ((err==0) && ((count == -1) || (vcount < count))) {
		thisobject=firstAvailableSlot();
		err=LoadObject(fp, thisobject, *index);
		if (err==0) {
			otype=objects[thisobject].object_type;
			CreateLoadedObject(thisobject, otype);
			SetLoadedResources(thisobject);
			cnum=objects[thisobject].num_children;
 
			(*index)++;
			
			if (((otype==XVForm) || (otype==XVBox)) && (cnum>0)) {
				LoadObjectTree(fp, thisobject, index, cnum);
				CurrentObject=toplevel;
				CurrentClass=objects[CurrentObject].pcore;
			}

			vcount++;
		}
	}

}

// --------------------------------------
// Read A String

int ReadString(FILE *fp, String datastring)
{
	char c;
	datastring[0]='\0';
	int ptr=0;

	c=fgetc(fp);
	while (c != '\n') {
		datastring[ptr++]=c;
		c=fgetc(fp);
	}
	datastring[ptr]='\0';
	return(ptr);
}		

// --------------------------------------
// Read string block

int readBlock(String buffer, int num, FILE *fp)
{
	int j, chr;

	for (j=0; j<MAXEDITBUFFERSIZE; j++)
		buffer[j]='\0';

	for (j=0; j<num; j++) {
		chr=fgetc(fp);	
		if (chr==EOF) break;			
		buffer[j]=chr;
	}
	buffer[j]='\0';
	return(j);
}
	
// --------------------------------------
// Load an object

int LoadObject(FILE *fp, int obj, int index)
{
	char dummy[60];
	int numres, err=0;
	
	loading_ref[index].reference=obj;

	ReadString(fp, dummy);

	if (strncmp(dummy, "--End-", 6)==0)  return(-1);

	ReadString(fp, objects[obj].name);
	ReadString(fp, objects[obj].obj_cursor);
	ReadString(fp, objects[obj].obj_font);
	ReadString(fp, objects[obj].obj_path);
	ReadString(fp, objects[obj].obj_icon);
	ReadString(fp, objects[obj].leftbitmap);
	ReadString(fp, objects[obj].obj_menu_font);
	ReadString(fp, objects[obj].background_colour);
	ReadString(fp, objects[obj].foreground_colour);
	ReadString(fp, objects[obj].border_colour);
	ReadString(fp, objects[obj].topshadow_colour);
	ReadString(fp, objects[obj].bottomshadow_colour);

	fscanf(fp, "%d %d\n", &objects[obj].object_type, &objects[obj].parent_type);

	ReadString(fp, loading_ref[index].index_name);
	ReadString(fp, loading_ref[index].window_name);
	fscanf(fp, "%d\n", &objects[obj].events);
	fscanf(fp, "%d\n", &objects[obj].num_children);
	fscanf(fp, "%d %d\n", &objects[obj].winxpos, &objects[obj].winypos);
	fscanf(fp, "%d %d\n", &objects[obj].winwidth, &objects[obj].winheight);
	
	int bool1, bool2, bool3;
	
	fscanf(fp, "%d %d\n", &bool1, &bool2); 
		objects[obj].fixed_size=bool1; 
		objects[obj].iconic=bool2;

	fscanf(fp, "%d\n", &bool1); 
		objects[obj].open_at_startup=bool1;

	fscanf(fp, "%d %d\n", &bool1, &bool2); 
		objects[obj].enabled=bool1; 
		objects[obj].visible=bool2;

	fscanf(fp, "%d %d %d\n", &bool1, &bool2, &bool3);
		objects[obj].ctrl=bool1; 
		objects[obj].echo=bool2;  
		objects[obj].autowrap=bool3;

	fscanf(fp, "%d %d %d %d %d %d %d\n", &objects[obj].vscroll_mode, &objects[obj].hscroll_mode, 
					&objects[obj].wrap_mode, &objects[obj].resize_mode, 
					&objects[obj].edit_mode, &objects[obj].shape,
					&objects[obj].justify);

	ReadString(fp, loading_ref[index].obref1);
	ReadString(fp, loading_ref[index].obref2);
	ReadString(fp, loading_ref[index].obref3);

	fscanf(fp, "%d\n", &objects[obj].object_time);
	fscanf(fp, "%d\n", &numres);
	for (int i=0; i<numres; i++) {
		ReadString(fp, resource_values[i].r_value);
	}

	int has;
	fscanf(fp, "%d %d %d\n", &has, &objects[obj].totalmenu, &objects[obj].nummenu);
	objects[obj].hasmenu=has;

	ReadString(fp, objects[obj].menu_background);
	ReadString(fp, objects[obj].menu_foreground);

	int len;
	char *buffer=new char[MAXEDITBUFFERSIZE];
	
	if (objects[obj].hasmenu) {

		if (objects[obj].nummenu > 0) 
			for (int i=0; i<objects[obj].nummenu; i++) 
				fscanf(fp, "%d\n", &objects[obj].numpulldown[i]);
				
		if (objects[obj].totalmenu > 0) {
			objects[obj].menuinfo=new struct menuentry[objects[obj].totalmenu];
			for (int i=0; i<objects[obj].totalmenu; i++) {
				fscanf(fp, "%d\n", &objects[obj].menuinfo[i].type);
				fscanf(fp, "%d\n", &objects[obj].menuinfo[i].listpoint);
				ReadString(fp, objects[obj].menuinfo[i].caption);
				ReadString(fp, objects[obj].menuinfo[i].name);
				ReadString(fp, objects[obj].menuinfo[i].callback);
				fscanf(fp, "%d\n", &len);
				if (len==0)
					objects[obj].menuinfo[i].codebuffer=NULL;
				else {
					objects[obj].menuinfo[i].codebuffer=new char[len+10];
					readBlock(buffer, len, fp);					
					strcpy(objects[obj].menuinfo[i].codebuffer, buffer);								
				}
			}
		}
	}

	for (int i=0; i<NUMEVENTS; i++) {
		fscanf(fp, "%d\n", &len);
		if (len != 0) {
			objects[obj].eventcode[i]=new char[len+10];
			readBlock(buffer, len, fp);
			strcpy(objects[obj].eventcode[i], buffer);								
		}
		else
			objects[obj].eventcode[i]=NULL;
	}

	for (int i=0; i<3; i++) {
		fscanf(fp, "%d\n", &len);
		if (len != 0) {
			objects[obj].callbackcode[i]=new char[len+10];
			readBlock(buffer, len, fp);
			strcpy(objects[obj].callbackcode[i], buffer);				
		}
		else
			objects[obj].callbackcode[i]=NULL;
	}

	delete buffer;
	ReadString(fp, dummy);
	return(err);
}

// -----------------------------------------
//

void CreateLoadedObject(int thisobject, objectType object)
{
	if (CurrentObject==-1) {
		objects[thisobject].parent_index=-1;
		objects[thisobject].parent_window=thisobject;
	}
	
	switch(object) {
		case XVTimer:
			objects[thisobject].ptimer = new xvTimer(objects[thisobject].name, 
				*CurrentClass, 0L, func_timer);			
			objects[thisobject].ptimer->SetDistance(0, 0);
			objects[thisobject].pcore->SetXY(0, 0);
			objects[thisobject].object_widget=objects[thisobject].ptimer->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(FALSE);
			break; 
		case XVList:
			objects[thisobject].plist = new xvVisualList(objects[thisobject].name, 
				*CurrentClass, func_dummy, 1, listvalues, 1);			
			objects[thisobject].plist->xvView::SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].plist->xvList::SetSize(100,100);
			objects[thisobject].plist->xvView::ManageChildren();
			objects[thisobject].object_widget=objects[thisobject].plist->xvView::ID();
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVButton:
			objects[thisobject].pbutton = new xvButton(objects[thisobject].name, *CurrentClass, func_dummy);			
			objects[thisobject].object_widget=objects[thisobject].pbutton->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].object_bitmap=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].leftbit=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVRepeater:
			objects[thisobject].prepeater = new xvRepeater(objects[thisobject].name, *CurrentClass, func_dummy, NULL, func_dummy, func_dummy);			
			objects[thisobject].object_widget=objects[thisobject].prepeater->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].object_bitmap=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].leftbit=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVBitmap:
			objects[thisobject].pbitmap = new xvBitmap(objects[thisobject].name, 
				*CurrentClass);			
			objects[thisobject].object_widget=objects[thisobject].pbitmap->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].object_bitmap=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			break;
		case XVLabel:
			objects[thisobject].plabel = new xvLabel(objects[thisobject].name, *CurrentClass);			
			objects[thisobject].object_widget=objects[thisobject].plabel->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].object_bitmap=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].leftbit=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVDropDown:
			objects[thisobject].plabel = new xvLabel(objects[thisobject].name, *CurrentClass);			
			objects[thisobject].object_widget=objects[thisobject].plabel->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVEditField:
			objects[thisobject].peditfield = new xvEditField(objects[thisobject].name, 
				*CurrentClass, XawtextRead, 25, 100, XawAsciiString, objects[thisobject].name);			
			objects[thisobject].object_widget=objects[thisobject].peditfield->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVBox:
			objects[thisobject].pbox = new xvBox(objects[thisobject].name, 
				*CurrentClass);			
			objects[thisobject].pbox->SetSize(100,100);
			objects[thisobject].object_widget=objects[thisobject].pbox->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			break;
		case XVToggle:
			objects[thisobject].ptoggle = new xvToggle(objects[thisobject].name, 
				*CurrentClass, func_dummy, objects[thisobject].name);			
			objects[thisobject].object_widget=objects[thisobject].ptoggle->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].object_bitmap=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].leftbit=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVForm:
			objects[thisobject].pform = new xvVisualForm(objects[thisobject].name, 
				*CurrentClass);			
			objects[thisobject].object_widget=objects[thisobject].pform->ID();
			objects[thisobject].pform->SetSize(100,100);
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			break;
		case XVVScroll:
			objects[thisobject].pvscroll = new xvScrollbar(objects[thisobject].name, 
				*CurrentClass, XtorientVertical, 200, 
				func_dummy, func_dummy, func_dummy);			
			objects[thisobject].object_widget=objects[thisobject].pvscroll->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			break;
		case XVHScroll:
			objects[thisobject].phscroll = new xvScrollbar(objects[thisobject].name, 
				*CurrentClass, XtorientHorizontal, 200, 
				func_dummy, func_dummy, func_dummy);			
			objects[thisobject].object_widget=objects[thisobject].phscroll->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			break;
		case XVPanner:
			objects[thisobject].ppanner = new xvPanner(objects[thisobject].name, 
				*CurrentClass, func_dummy, 100, 100, 50, 50);			
			objects[thisobject].ppanner->SetSize(100,100);
			objects[thisobject].object_widget=objects[thisobject].ppanner->ID();
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			break;
		case XVFileList:
			objects[thisobject].pfilelist= new xvFileList(objects[thisobject].name, 
				*CurrentClass, objects[thisobject].obj_path, XVFILELIST, 
				FALSE, func_dummy, FALSE);
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].pfilelist->xvList::SetSize(150,180);
			objects[thisobject].pfilelist->xvView::ManageChildren();
			objects[thisobject].object_widget=objects[thisobject].pfilelist->xvView::ID();
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;
		case XVDirList:
			objects[thisobject].pdirlist= new xvFileList(objects[thisobject].name, 
				*CurrentClass, objects[thisobject].obj_path, XVDIRLIST, 
				FALSE, func_dummy, FALSE);
			objects[thisobject].pcore->SetFixedPosition(TRUE);
			objects[thisobject].pcore->SetResizable(TRUE);
			objects[thisobject].pdirlist->xvList::SetSize(150,180);
			objects[thisobject].pdirlist->xvView::ManageChildren();
			objects[thisobject].object_widget=objects[thisobject].pdirlist->xvView::ID();
			objects[thisobject].object_font= new xvFont(*objects[thisobject].pcore, 
				objects[thisobject].obj_font);
			objects[thisobject].pcore->SetResource(XtNfont, 
				(XtArgVal) objects[thisobject].object_font->FontRef());
			break;  
		case XVWindow:
			objects[thisobject].pwindow = new xvWindow(objects[thisobject].name,
					objects[thisobject].name , xvisContext);
			objects[thisobject].pwindow->SetSize(300, 300);				
		
			objects[thisobject].pwindow->SetWindowResource(
				XtNgeometry, (XtArgVal)"+310+260");
			objects[thisobject].pwindow->Open();
			objects[thisobject].object_widget=objects[thisobject].pwindow->ID();
			objects[thisobject].pwindow->GCSetForeground(7);
			objects[thisobject].object_bitmap=new xvBitmap(
				objects[thisobject].name, *objects[thisobject].pcore, FALSE);
		
			objects[thisobject].pwindow->AddEvent(StructureNotifyMask,
				(XtEventHandler) WindowEvent,
				(XtPointer) objects[thisobject].name);

			objects[thisobject].pwindow->AddEvent(EnterWindowMask,
				(XtEventHandler) enter_call,
				(XtPointer) objects[thisobject].name);

			objects[thisobject].pwindow->AddEvent(LeaveWindowMask,
				(XtEventHandler) leave_call,
				(XtPointer) objects[thisobject].name);

			objects[thisobject].pwindow->AddEvent(ExposureMask,
				(XtEventHandler) winexpose,
				(XtPointer) objects[thisobject].name);

			drawGrid(thisobject);
			break;
		default:
			break;
	}
	
	objects[thisobject].pcore->AddEvent(ButtonPressMask, 
			(XtEventHandler) ObjectButtonEvent, 
			(XtPointer) objects[thisobject].name);
		
	if (uses3D[object])
		objects[thisobject].pcore->SetBeNiceToColormap(FALSE);

	objectInstance[object]++;
	nobject++;
}

// ------------------------------------------
// Resource Settings

void SetLoadedResources(int obj)
{
	objectType otype=objects[obj].object_type;
	int numres=object_resource_number[otype];
	char* *types=object_resource_types[otype];
	resourceType *datatypes=object_data_types[otype];

	prop_object_index=obj;
	prop_object_type=otype;
	for (int i=0; i<numres; i++) {
		resource_values[i].r_restype=types[i];
		resource_values[i].r_type=datatypes[i];
		strcpy(propeditbuffer, resource_values[i].r_value);
		selected_index=i;
		objects[obj].pcore->Unmap();
		setResourceExec(FALSE);		
		objects[obj].pcore->Map();
	}

}

// ------------------------------------------
// Relink inter- object references

void RelinkReferences(int index)
{
	int thisobject;
	for (int i=0; i<index; i++) {
		thisobject=loading_ref[i].reference;
		if (strcmp(loading_ref[i].index_name, "(None)")==0)
			objects[thisobject].parent_index=-1;
		else
			objects[thisobject].parent_index=ObjectByName(loading_ref[i].index_name);

		objects[thisobject].parent_window=ObjectByName(loading_ref[i].window_name);

		if (strcmp(loading_ref[i].obref1, "None")==0)
			objects[thisobject].object_ref1=-1;
		else
			objects[thisobject].object_ref1=ObjectByName(loading_ref[i].obref1);

		if (strcmp(loading_ref[i].obref2, "None")==0)
			objects[thisobject].object_ref2=-1;
		else
			objects[thisobject].object_ref2=ObjectByName(loading_ref[i].obref2);

		if (strcmp(loading_ref[i].obref3, "None")==0)
			objects[thisobject].object_ref3=-1;
		else
			objects[thisobject].object_ref3=ObjectByName(loading_ref[i].obref3);
	}
}

// ---------------------------------------
// Load project

int LoadProject()
{
	char tempbuffer[180];
	int num=0, mnum=0;
	FILE *pp;
	pp=fopen(project_path, "r");
	if (pp!=NULL) {
		fscanf(pp, "%d\n", &num);
		for (int i=0; i<num; i++) {
			ReadString(pp, tempbuffer);
			LoadCurrentForm(tempbuffer);
		}
		fscanf(pp, "%d\n", &mnum);
		for (int i=0; i<mnum; i++) {
			ReadString(pp, tempbuffer);
			loadCurrentModule(tempbuffer);	
		}
		fclose(pp);
	}			
	updateProjectListWindow();		

}
	