
/*     *******************************************************************************
	XVisual GUI Development System for Linux/X-Window
	Copyright (C) 1996 Steve Carrie.

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
        ******************************************************************************* */

/* =============================================
	xvis user interface builder for Athena widgets using
	xv class library.

	xvisMenus segment

	V1.0BETA  - last edited 25/06/96
    ============================================= */

// ---------------------------------------------
// Local Include file declarations

#include "xvis.h"
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "bitmaps/upptr"
#include "bitmaps/downptr"
#include "bitmaps/leftptr"
#include "bitmaps/rightptr"

#define MAXMENU	100
#define XVMENUOK		1
#define XVMENUCANCEL	2

#define XVMENUNOERR	0
#define XVMENUNOMAIN	1
#define XVMENUNOSUB	2
#define XVMENUERROR	3

// ---------------------------------------------
// Local function declarations

void makeMenuBar(int targ, int num, char **args);
int parseMenuData();
void defaultMenuWindow();
void closeMenuWindow(int mode);
int findEntry(String caption, String name, String callback);
void dataAlert(String header, String info);
void updateListInfo();
void updateListEntry();
void insertListEntry(int etype);
int findVoidEntry();
int findListPosition(int pos);
void buildList();
void readMenu();
void recoverEditInfo();

void alert_cancel(Widget aw, XtPointer client, XtPointer call);
void func_menu_ok(Widget aw, XtPointer client, XtPointer call);
void func_menu_cancel(Widget aw, XtPointer client, XtPointer call);
void add_callback(Widget aw, XtPointer client, XtPointer call);
void remove_callback(Widget aw, XtPointer client, XtPointer call);
void clear_callback(Widget aw, XtPointer client, XtPointer call);
void up_callback(Widget aw, XtPointer client, XtPointer call);
void down_callback(Widget aw, XtPointer client, XtPointer call);
void main_callback(Widget aw, XtPointer client, XtPointer call);
void sub_callback(Widget aw, XtPointer client, XtPointer call);
void list_callback(Widget aw, XtPointer client, XtPointer call);

// ---------------------------------------------
// Local Data

int target=-1;

int numeditlist=0;
char **menueditlist;
struct menuentry *editlist;
int menuchecklist[MAXMENU];

int selected_entry=-1;
char selected_string[40];

char menu_caption[40];
char menu_name[40];
char menu_callback[40];

char *menu_build[40];

int nummain;
int numsub[30];

XtCallbackProc dummy_list[]={	func_dummy, func_dummy, func_dummy, func_dummy, 
							func_dummy, func_dummy, func_dummy, func_dummy, 
							func_dummy, func_dummy, func_dummy, func_dummy, 
							func_dummy, func_dummy, func_dummy, func_dummy, 
							func_dummy, func_dummy, func_dummy, func_dummy };

// ---------------------------------------------
// Local callbacks

void dataAlert(String header, String info)
{
	alert.init(header, xvisControl, info,
		XVMSGSIMPLEFATAL, "red",
		alert_cancel, alert_cancel);
	alert.Open();
}

void alert_cancel(Widget aw, XtPointer client, XtPointer call)
{
	alert.Close();
}

void func_menu_cancel(Widget aw, XtPointer client, XtPointer call)
{
	closeMenuWindow(XVMENUCANCEL);
}

void func_menu_ok(Widget aw, XtPointer client, XtPointer call)
{
	int ok=parseMenuData();

	if ((ok==XVMENUNOERR) || (numeditlist<=1)) 
		closeMenuWindow(XVMENUOK);
	else 
		switch(ok) {
			case XVMENUNOMAIN:
				dataAlert("Parse Error!", "No main object for sub-object(s)!");	
				break;
			case XVMENUNOSUB:
				dataAlert("Parse Error!", "No sub object(s) for main object");	
				break;
			default:
				dataAlert("Parse Error!", "Your menu contains one or more \n errors!");	
				break;
		}
}

void update_callback(Widget aw, XtPointer client, XtPointer call)
{
	recoverEditInfo();
	if ( (strlen(menu_caption)!=0) && (strlen(menu_name) !=0) 
		&& (strlen(menu_callback)!=0)) {

		updateListEntry();
		menuCaption.SetResource(XtNstring, (XtArgVal) NULL);		
		menuName.SetResource(XtNstring, (XtArgVal) NULL);		
		menuCallback.SetResource(XtNstring, (XtArgVal) NULL);		
		menu_caption[0]='\0';
		menu_name[0]='\0';
		menu_callback[0]='\0';
			
	}
	else 
		dataAlert("Missing Information", "Menu entry name and/or\nCaption and/or callback missing!");	
}

void addsep_callback(Widget aw, XtPointer client, XtPointer call)
{
	strcpy(menu_caption, "---");
	strcpy(menu_name, "seperator");
	strcpy(menu_callback, "_app_dummy_callback");
	insertListEntry(XVMENUSEP);
	menuCaption.SetResource(XtNstring, (XtArgVal) NULL);		
	menuName.SetResource(XtNstring, (XtArgVal) NULL);		
	menuCallback.SetResource(XtNstring, (XtArgVal) NULL);		
	menu_caption[0]='\0';
	menu_name[0]='\0';
	menu_callback[0]='\0';
}

void add_callback(Widget aw, XtPointer client, XtPointer call)
{
	recoverEditInfo();
	if ( (strlen(menu_caption)!=0) && (strlen(menu_name) !=0) 
		&& (strlen(menu_callback)!=0)) {

		if (findEntry(menu_caption, NULL, NULL)!=-1) {
			dataAlert("Duplicate Information", "Duplicate Menu Caption!");	
			return;
		}		

		if(findEntry(NULL, menu_name, NULL)!=-1)	{
			dataAlert("Duplicate Information", "Duplicate Menu object name!");	
			return;
		}		

		if(findEntry(NULL, NULL, menu_callback)!=-1)	{
			dataAlert("Duplicate Information", "Duplicate Menu object callback!");	
			return;
		}		

		insertListEntry(XVMENUMAIN);
		menuCaption.SetResource(XtNstring, (XtArgVal) NULL);		
		menuName.SetResource(XtNstring, (XtArgVal) NULL);		
		menuCallback.SetResource(XtNstring, (XtArgVal) NULL);		
		menu_caption[0]='\0';
		menu_name[0]='\0';
		menu_callback[0]='\0';
			
	}
	else 
		dataAlert("Missing Information", "Menu entry name and/or\nCaption and/or callback missing!");	
	
}

void remove_callback(Widget aw, XtPointer client, XtPointer call)
{
	int loc=0, pos=0;

	updateListInfo();
	if (selected_entry==-1) {
		dataAlert("REMOVE", "No Object Selected!");	
		return;
	}	
	loc=menuchecklist[selected_entry];
	if (editlist[loc].type != XVMENUEND) {
		pos=editlist[loc].listpoint;
		editlist[loc].type=XVMENUVOID;
		if (editlist[loc].codebuffer != NULL)
			delete editlist[loc].codebuffer;
		for (int i=0; i<MAXMENU; i++)
			if ((editlist[i].type != XVMENUVOID) && (editlist[i].listpoint > pos)) 
				editlist[i].listpoint--;		
		buildList();
	}
}

void clear_yes(Widget aw, XtPointer client, XtPointer call)
{
	alert.Close();
	defaultMenuWindow();
}

void clear_no(Widget aw, XtPointer client, XtPointer call)
{
	alert.Close();
}

void clear_callback(Widget aw, XtPointer client, XtPointer call)
{
	if (numeditlist >1) {
		alert.init("Clear Menu", xvisControl, 
			"This will remove all menu entries.\nAre you sure you want to do this?",
			XVMSGSIMPLECHOICE, "green",
			clear_yes, clear_no);
		alert.Open();
	}
	else
		defaultMenuWindow();

}

void up_callback(Widget aw, XtPointer client, XtPointer call)
{
	int newpos=-1;

	updateListInfo();
	if (selected_entry==-1) {
		dataAlert("Shift Up", "No Object Selected!");	
		return;
	}	
	int loc=menuchecklist[selected_entry];
	if (editlist[loc].type!=XVMENUEND) {
		if (editlist[loc].listpoint>0) {
			int np=editlist[loc].listpoint;
			newpos=findListPosition(np-1);
			editlist[loc].listpoint=editlist[newpos].listpoint;
			editlist[newpos].listpoint=np;
			buildList();	
			menuList.Highlight(editlist[loc].listpoint);
		}
	}
	else 
		dataAlert("Move Up", "Cannot move this object!");		
}

void down_callback(Widget aw, XtPointer client, XtPointer call)
{
	int newpos=-1, np;

	updateListInfo();
	if (selected_entry==-1) {
		dataAlert("Shift Down", "No Object Selected!");	
		return;
	}	
	int loc=menuchecklist[selected_entry];
	if (editlist[loc].type!=XVMENUEND) {
		np=editlist[loc].listpoint;
		if (np<numeditlist-2) {
			newpos=findListPosition(np+1);
			editlist[loc].listpoint=editlist[newpos].listpoint;
			editlist[newpos].listpoint=np;
			buildList();	
			menuList.Highlight(editlist[loc].listpoint);
		}
	}
	else 
		dataAlert("Move Down", "Cannot move this object!");		
}

void main_callback(Widget aw, XtPointer client, XtPointer call)
{
	char tempbuffer[40], i;

	updateListInfo();
	if (selected_entry==-1) {
		dataAlert("Set Main", "No Object Selected!");	
		return;
	}	
	int loc=menuchecklist[selected_entry];
	if (editlist[loc].type==XVMENUSUB) {
		editlist[loc].type=XVMENUMAIN;
		for (i=5; i<strlen(editlist[loc].caption); i++)
			tempbuffer[i-5]=editlist[loc].caption[i];
		tempbuffer[i-5]='\0';
		strcpy(editlist[loc].caption, tempbuffer);
		buildList();
		menuEditCallback.SetSensitive(FALSE);
	}
	else 
		dataAlert("Set Main", "Cannot set this object to be\na main menu object!");		

	menuList.Highlight(editlist[loc].listpoint);
}

void sub_callback(Widget aw, XtPointer client, XtPointer call)
{
	char tempbuffer[40];

	updateListInfo();
	if (selected_entry==-1) {
		dataAlert("Set Sub", "No Object Selected!");	
		return;
	}	
	int loc=menuchecklist[selected_entry];
	if (editlist[loc].type==XVMENUMAIN) {
		editlist[loc].type=XVMENUSUB;
		strcpy(tempbuffer, "-----");
		strcat(tempbuffer, editlist[loc].caption);
		strcpy(editlist[loc].caption, tempbuffer);
		buildList();
		menuEditCallback.SetSensitive(TRUE);
	}
	else 
		dataAlert("Set Sub", "Cannot set this object to be\na sub-menu object!");		

	menuList.Highlight(editlist[loc].listpoint);
}

void updateListInfo()
{
	XawListReturnStruct *val=XawListShowCurrent(menuList.ID());
	selected_entry=val->list_index;
	strcpy(selected_string, val->string);
	if ((selected_entry==XAW_LIST_NONE) || (selected_entry>MAXMENU)) {
		selected_entry=-1;
		selected_string[0]='\0';
	}
}

void list_callback(Widget aw, XtPointer client, XtPointer call)
{
	int loc=-1;
	updateListInfo();
	menuEditCallback.SetSensitive(FALSE);
	if (selected_entry!=-1) {
		loc=menuchecklist[selected_entry];
		if ((editlist[loc].type != XVMENUEND) && (editlist[loc].type != XVMENUSEP)) {
			strcpy(menu_caption, editlist[loc].caption);
			strcpy(menu_name, editlist[loc].name);
			strcpy(menu_callback, editlist[loc].callback);
			menuCaption.SetResource(XtNstring, (XtArgVal) menu_caption);		
			menuName.SetResource(XtNstring, (XtArgVal) menu_name);		
			menuCallback.SetResource(XtNstring, (XtArgVal) menu_callback);		
			if (editlist[loc].type==XVMENUSUB)
				menuEditCallback.SetSensitive(TRUE);
		}
		else {
			menuCaption.SetResource(XtNstring, (XtArgVal) NULL);		
			menuName.SetResource(XtNstring, (XtArgVal) NULL);		
			menuCallback.SetResource(XtNstring, (XtArgVal) NULL);		
			menu_caption[0]='\0';
			menu_name[0]='\0';
			menu_callback[0]='\0';
		}
	}
	else 
		if (selected_entry==numeditlist) {
			menuCaption.SetResource(XtNstring, (XtArgVal) NULL);		
			menuName.SetResource(XtNstring, (XtArgVal) NULL);		
			menuCallback.SetResource(XtNstring, (XtArgVal) NULL);		
			menu_caption[0]='\0';
			menu_name[0]='\0';
			menu_callback[0]='\0';
		}
}

void menu_font_ok(Widget aw, XtPointer client, XtPointer call)
{
	xvisFonts.Close();
	strcpy(objects[target].obj_menu_font, propeditbuffer);
}

void font_callback(Widget aw, XtPointer client, XtPointer call)
{
	buildFontsWindow(menu_font_ok);
	xvisFonts.Open();
}

void menu_back_ok(Widget aw, XtPointer client, XtPointer call)
{
	xvisColours.Close();
	strcpy(objects[target].menu_background, propeditbuffer);
}

void menu_back(Widget aw, XtPointer client, XtPointer call)
{
	buildColourWindow(menu_back_ok);
	xvisColours.Open();
}

void menu_fore_ok(Widget aw, XtPointer client, XtPointer call)
{
	xvisColours.Close();
	strcpy(objects[target].menu_foreground, propeditbuffer);
}

void menu_fore(Widget aw, XtPointer client, XtPointer call)
{
	buildColourWindow(menu_fore_ok);
	xvisColours.Open();
}

void menu_func_ok(Widget aw, XtPointer client, XtPointer call)
{
	char *temp=(char *) client;
	int which;
	sscanf(temp, "%d", &which);
	int obj=editor[which].object;
	int ref=editor[which].ref;

	XtVaGetValues(editor[which].editor->xvEditField::ID(), XtNstring, &temp, NULL);
	if (editlist[obj].codebuffer != NULL) {
		delete editlist[obj].codebuffer;
		editlist[obj].codebuffer=NULL;
	}
	editlist[obj].codebuffer=new char[strlen(temp)+10];
	strcpy(editlist[obj].codebuffer, temp);

	delete editor[which].editor;
	delete editor[which].edit_buffer;
	editor[which].in_use=FALSE;
		
}

void menu_func_cancel(Widget aw, XtPointer client, XtPointer call)
{
	char *temp=(char *) client;
	int which;
	sscanf(temp, "%d", &which);
	delete editor[which].editor;
	delete editor[which].edit_buffer;
	editor[which].in_use=FALSE;
}

void menu_editcall(Widget aw, XtPointer client, XtPointer call)
{
	updateListInfo();
	if (selected_entry==-1) {
		dataAlert("Edit Function", "No Object Selected!");	
		return;
	}	

	int loc=menuchecklist[selected_entry];
	int ptr=findFirstEditEntry();
	if (ptr==-1) {
		dataAlert("Edit Function", "Can't Edit - Out of Space!");	
		return;
	}	

	editor[ptr].in_use=TRUE;
	editor[ptr].object=loc;
	editor[ptr].mode=0;
	editor[ptr].ref=selected_entry;
	sprintf(editor[ptr].client, "%d", ptr);
	sprintf(editor[ptr].name, "Menu function %s", editlist[loc].callback);
	editor[ptr].edit_buffer=new char[MAXEDITBUFFERSIZE];
	if (editlist[loc].codebuffer==NULL) 
		sprintf(editor[ptr].edit_buffer, 
			"void %s (Widget aw, XtPointer clientdata, XtPointer calldata)\n{\n\n}\n", 
			editlist[loc].callback);
	else
		strcpy(editor[ptr].edit_buffer, editlist[loc].codebuffer);

	editor[ptr].editor=new xvEditWindow( xvisContext, editor[ptr].name, 
		XawtextEdit, editor[ptr].edit_buffer, 600, 400, 
		menu_func_ok, menu_func_cancel, editor[ptr].client);
	editor[ptr].editor->xvEditField::ChangeFont("*times-*-r-*-140*");		

	editor[ptr].editor->Open();
}

// ---------------------------------------------
// Local Events

void Caption_entry(Widget w, XEvent *event, String *params, Cardinal *num)
{
	XawTextPosition start, end;
	XawTextBlock block;

	menuCaption.Position(&start, &end);
	menuCaption.SourceRead(0, &block, 40);
	strncpy(menu_caption, block.ptr, end);
	menu_caption[end]='\0';
	menuCaption.UnsetSelection();
	if ((strlen(menu_caption)>0) && (strlen(menu_name)==0)) {
		strcpy(menu_name,menu_caption);
		strcat(menu_name, "_menu");
		for (int i=0; i<strlen(menu_name); i++) {
			menu_name[i]=tolower(menu_name[i]);
			if (menu_name[i]==' ') menu_name[i]='_';
		}
		menuName.SetResource(XtNstring, (XtArgVal) menu_name);	
	}
	xvisMenus.MovePointer(150, 60);
}

void Name_entry(Widget w, XEvent *event, String *params, Cardinal *num)
{
	XawTextPosition start, end;
	XawTextBlock block;

	menuName.Position(&start, &end);
	menuName.SourceRead(0, &block, 40);
	strncpy(menu_name, block.ptr, end);
	menu_name[end]='\0';
	menuName.UnsetSelection();
	if ((strlen(menu_name)>0) && (strlen(menu_callback)==0)) {
		strcpy(menu_callback, menu_name);
		strcat(menu_callback, "_select");
		menuCallback.SetResource(XtNstring, (XtArgVal) menu_callback);	
	}
	xvisMenus.MovePointer(150, 90);
}

void Callback_entry(Widget w, XEvent *event, String *params, Cardinal *num)
{
	XawTextPosition start, end;
	XawTextBlock block;

	menuCallback.Position(&start, &end);
	menuCallback.SourceRead(0, &block, 40);
	strncpy(menu_callback, block.ptr, end);
	menu_callback[end]='\0';
	menuCallback.UnsetSelection();
	xvisMenus.MovePointer(150, 30);
	add_callback(w, NULL, NULL);

}

void recoverEditInfo()
{
	char *s;
	XtVaGetValues(menuCaption.ID(), XtNstring, &s, NULL);
	strcpy(menu_caption, s);
	XtVaGetValues(menuName.ID(), XtNstring, &s, NULL);
	strcpy(menu_name, s);
	XtVaGetValues(menuCallback.ID(), XtNstring, &s, NULL);
	strcpy(menu_callback, s);
}

// ---------------------------------------------
// Local functions
// Build the menu window

void buildMenuWindow()
{
	xvisMenus.init("Menu Builder", "xvisMenus", xvisContext);
	xvisMenus.SetBackground("gray70");
	xvisMenus.SetWindowResource(XtNgeometry, (XtArgVal)"+200+220");
 	xvisMenus.SetWindowResource(XtNiconPixmap, 
		xvisMenus.SetPixmapInfo(xvis_xpm, NULL));

	menuCaptionLabel.init("Menu Caption", xvisMenus);
	menuCaptionLabel.SetBackground("gray70");
	menuCaptionLabel.SetForeground("blue");
	menuCaptionLabel.SetDistance(10, 20);
	menuCaptionLabel.ChangeFont("*times-*-r-*-140*"); 
	menuCaptionLabel.SetBorderWidth(0);
	menuCaptionLabel.SetFixedPosition(TRUE);

	menuCaption.init("menuCaption", xvisMenus, XawtextEdit, 20, 150);
	menuCaption.SetDistance(130, 20);
	XtTranslations caption_trans=XtParseTranslationTable(
		"<Key>Return: select-all()  caption_entry() \n\
		!Meta<Key> i: no-op()");
	menuCaption.OverrideTranslations(caption_trans);
	xvisContext.AddAction("caption_entry", (XtActionProc) Caption_entry);
	menuCaption.SetFixedPosition(TRUE);

	menuNameLabel.init("Object Name", xvisMenus);
	menuNameLabel.SetBackground("gray70");
	menuNameLabel.SetForeground("blue");
	menuNameLabel.SetDistance(10, 50);
	menuNameLabel.ChangeFont("*times-*-r-*-140*"); 
	menuNameLabel.SetBorderWidth(0);
	menuNameLabel.SetFixedPosition(TRUE);

	menuName.init("menuName", xvisMenus, XawtextEdit, 20, 150);
	menuName.SetDistance(130, 50);
	XtTranslations name_trans=XtParseTranslationTable(
		"<Key>Return: select-all()  name_entry() \n\
		!Meta<Key> i: no-op()");
	menuName.OverrideTranslations(name_trans);
	xvisContext.AddAction("name_entry", (XtActionProc) Name_entry);
	menuName.SetFixedPosition(TRUE);

	menuCallbackLabel.init("Object Function", xvisMenus);
	menuCallbackLabel.SetBackground("gray70");
	menuCallbackLabel.SetForeground("blue");
	menuCallbackLabel.SetDistance(10, 80);
	menuCallbackLabel.ChangeFont("*times-*-r-*-140*"); 
	menuCallbackLabel.SetBorderWidth(0);
	menuCallbackLabel.SetFixedPosition(TRUE);

	menuCallback.init("menuCallback", xvisMenus, XawtextEdit, 20, 150);
	menuCallback.SetDistance(130, 80);
	XtTranslations callback_trans=XtParseTranslationTable(
		"<Key>Return: select-all()  callback_entry() \n\
		!Meta<Key> i: no-op()");
	menuCallback.OverrideTranslations(callback_trans);
	xvisContext.AddAction("callback_entry", (XtActionProc) Callback_entry);
	menuCallback.SetFixedPosition(TRUE);

	menuOK.init("OK", xvisMenus, func_menu_ok);
	menuOK.SetDistance(400, 20);
	menuOK.ChangeFont("*times-*-r-*-140*"); 
	menuOK.SetWidth(75);
	menuOK.SetShadowWidth(4);
	menuOK.SetBackground("gray85");
	menuOK.SetFixedPosition(TRUE);
	menuOK.SetBeNiceToColormap(FALSE);

	menuCancel.init("Cancel", xvisMenus, func_menu_cancel);
	menuCancel.SetDistance(400, 60);
	menuCancel.ChangeFont("*times-*-r-*-140*"); 
	menuCancel.SetWidth(75);
	menuCancel.SetShadowWidth(4);
	menuCancel.SetBackground("gray85");
	menuCancel.SetFixedPosition(TRUE);
	menuCancel.SetBeNiceToColormap(FALSE);

	menuFont.init("Font", xvisMenus, font_callback);
	menuFont.SetDistance(400, 120);
	menuFont.ChangeFont("*times-*-r-*-140*"); 
	menuFont.SetWidth(75);
	menuFont.SetHeight(50);
	menuFont.SetShadowWidth(4);
	menuFont.SetBackground("gray85");
	menuFont.SetFixedPosition(TRUE);
	menuFont.SetBeNiceToColormap(FALSE);

	menuEditCallback.init("Edit Func.", xvisMenus, menu_editcall);
	menuEditCallback.SetDistance(300, 80);
	menuEditCallback.ChangeFont("*times-*-r-*-140*"); 
	menuEditCallback.SetWidth(90);
	menuEditCallback.SetHeight(20);
	menuEditCallback.SetShadowWidth(3);
	menuEditCallback.SetBackground("gray85");
	menuEditCallback.SetFixedPosition(TRUE);
	menuEditCallback.SetSensitive(FALSE);
	menuEditCallback.SetBeNiceToColormap(FALSE);

	menuBack.init("Background", xvisMenus, menu_back);
	menuBack.SetDistance(300, 120);
	menuBack.ChangeFont("*times-*-r-*-140*"); 
	menuBack.SetWidth(90);
	menuBack.SetHeight(20);
	menuBack.SetShadowWidth(3);
	menuBack.SetBackground("gray85");
	menuBack.SetFixedPosition(TRUE);
	menuBack.SetBeNiceToColormap(FALSE);

	menuFore.init("Foreground", xvisMenus, menu_fore);
	menuFore.SetDistance(300, 150);
	menuFore.ChangeFont("*times-*-r-*-140*"); 
	menuFore.SetWidth(90);
	menuFore.SetHeight(20);
	menuFore.SetShadowWidth(3);
	menuFore.SetBackground("gray85");
	menuFore.SetFixedPosition(TRUE);
	menuFore.SetBeNiceToColormap(FALSE);

	menuDown.init("Down", xvisMenus, down_callback);
	menuDown.SetDistance(40, 140);
	menuDown.SetBitmapInfo(downptr_bits, downptr_width, downptr_height, XtNbitmap);
	menuDown.SetShadowWidth(4);
	menuDown.SetFixedPosition(TRUE);
	menuDown.SetBackground("gray85");
	menuDown.SetBeNiceToColormap(FALSE);

	menuUp.init("Up", xvisMenus, up_callback);
	menuUp.SetDistance(80, 140);
	menuUp.SetBitmapInfo(upptr_bits, upptr_width, upptr_height, XtNbitmap);
	menuUp.SetShadowWidth(4);
	menuUp.SetBackground("gray85");
	menuUp.SetFixedPosition(TRUE);
	menuUp.SetBeNiceToColormap(FALSE);

	menuMain.init("Main", xvisMenus, main_callback);
	menuMain.SetDistance(120, 140);
	menuMain.SetBitmapInfo(leftptr_bits, leftptr_width, leftptr_height, XtNbitmap);
	menuMain.SetShadowWidth(4);
	menuMain.SetBackground("gray85");
	menuMain.SetFixedPosition(TRUE);
	menuMain.SetBeNiceToColormap(FALSE);

	menuSub.init("Sub", xvisMenus, sub_callback);
	menuSub.SetDistance(160, 140);
	menuSub.SetBitmapInfo(rightptr_bits, rightptr_width, rightptr_height, XtNbitmap);
	menuSub.SetShadowWidth(4);
	menuSub.SetBackground("gray85");
	menuSub.SetFixedPosition(TRUE);
	menuSub.SetBeNiceToColormap(FALSE);
	
	menuView.init("menuView", xvisMenus, TRUE, FALSE, TRUE);
	menuView.SetDistance(20, 180);
	menuView.SetShadowWidth(4);
	menuView.SetFixedPosition(TRUE);

	menuList.init("menuList", menuView, list_callback, numeditlist, menueditlist, 1);
	menuView.SetSize(360, 200);
	menuList.SetFixedPosition(TRUE);

	menuAdd.init("Add", xvisMenus, add_callback);
	menuAdd.SetDistance(400, 190);
	menuAdd.ChangeFont("*times-*-r-*-140*"); 
	menuAdd.SetWidth(75);
	menuAdd.SetShadowWidth(4);
	menuAdd.SetBackground("gray85");
	menuAdd.SetFixedPosition(TRUE);
	menuAdd.SetBeNiceToColormap(FALSE);

	menuAddSep.init("Add Sep.", xvisMenus, addsep_callback);
	menuAddSep.SetDistance(400, 230);
	menuAddSep.ChangeFont("*times-*-r-*-140*"); 
	menuAddSep.SetWidth(75);
	menuAddSep.SetShadowWidth(4);
	menuAddSep.SetBackground("gray85");
	menuAddSep.SetFixedPosition(TRUE);
	menuAddSep.SetBeNiceToColormap(FALSE);

	menuUpdate.init("Update", xvisMenus, update_callback);
	menuUpdate.SetDistance(400, 270);
	menuUpdate.ChangeFont("*times-*-r-*-140*"); 
	menuUpdate.SetWidth(75);
	menuUpdate.SetShadowWidth(4);
	menuUpdate.SetBackground("gray85");
	menuUpdate.SetFixedPosition(TRUE);
	menuUpdate.SetBeNiceToColormap(FALSE);

	menuRemove.init("Remove", xvisMenus, remove_callback);
	menuRemove.SetDistance(400, 310);
	menuRemove.ChangeFont("*times-*-r-*-140*"); 
	menuRemove.SetWidth(75);
	menuRemove.SetShadowWidth(4);
	menuRemove.SetBackground("gray85");
	menuRemove.SetFixedPosition(TRUE);
	menuRemove.SetBeNiceToColormap(FALSE);

	menuClear.init("Clear", xvisMenus, clear_callback);
	menuClear.SetDistance(400, 350);
	menuClear.ChangeFont("*times-*-r-*-140*"); 
	menuClear.SetWidth(75);
	menuClear.SetShadowWidth(4);
	menuClear.SetBackground("gray85");
	menuClear.SetFixedPosition(TRUE);
	menuClear.SetBeNiceToColormap(FALSE);

	MenusVisible=FALSE;
}

// ---------------------------------------------
// Init the menu window

void defaultMenuWindow()
{
	numeditlist=0;

	menueditlist=new char*[MAXMENU];
	editlist=new struct menuentry[MAXMENU];
	
	for (int i=0; i<MAXMENU; i++) {
		editlist[i].type=XVMENUVOID;
		editlist[i].name[0]='\0';
		editlist[i].callback[0]='\0';
		editlist[i].caption[0]='\0';
		editlist[i].listpoint=-1;
		editlist[i].codebuffer=NULL;
		menueditlist[i]=NULL;
	}
	editlist[0].type=XVMENUEND;
	editlist[0].listpoint=0;
	strcpy(editlist[0].caption, "--End of list--");
	buildList();
	menuCaption.SetResource(XtNstring, (XtArgVal) NULL);		
	menuName.SetResource(XtNstring, (XtArgVal) NULL);		
	menuCallback.SetResource(XtNstring, (XtArgVal) NULL);		
}

void buildList()
{
	int ptr=0;

	numeditlist=0;
	for (int i=0; i<MAXMENU; i++)
		if (editlist[i].type != XVMENUVOID) {
			ptr=editlist[i].listpoint;
			menueditlist[ptr]=editlist[i].caption;
			menuchecklist[ptr]=i;
			numeditlist++;
		}
	menuList.SetList(numeditlist, menueditlist);
	menuView.SetSize(360, 200);
}

void readMenu()
{
	char tempbuffer[40];

	defaultMenuWindow();
	numeditlist=objects[target].totalmenu;
		
	for (int i=0; i<numeditlist; i++) {
		editlist[i].type=objects[target].menuinfo[i].type;
		editlist[i].listpoint=objects[target].menuinfo[i].listpoint;
		strcpy(editlist[i].name, objects[target].menuinfo[i].name);
		strcpy(editlist[i].callback, objects[target].menuinfo[i].callback);
		strcpy(tempbuffer, objects[target].menuinfo[i].caption);
		if (editlist[i].type==XVMENUSUB) {
			strcpy(tempbuffer, "-----");
			strcat(tempbuffer, objects[target].menuinfo[i].caption);
		}	
		strcpy(editlist[i].caption, tempbuffer);
		editlist[i].codebuffer=NULL;
		if (objects[target].menuinfo[i].codebuffer != NULL) {
			editlist[i].codebuffer=new char[strlen(objects[target].menuinfo[i].codebuffer)+10];
			strcpy(editlist[i].codebuffer, objects[target].menuinfo[i].codebuffer);
		}
	}
	editlist[numeditlist].type=XVMENUEND;
	editlist[numeditlist].listpoint=numeditlist;
	strcpy(editlist[numeditlist].caption, "--End of list--");
	numeditlist++;
	buildList();
	menuCaption.SetResource(XtNstring, (XtArgVal) NULL);		
	menuName.SetResource(XtNstring, (XtArgVal) NULL);		
	menuCallback.SetResource(XtNstring, (XtArgVal) NULL);		
}

// --------------------------------------------
// Open window & init

void openMenuWindow()
{
	target=objects[CurrentObject].parent_window;
	menuEditCallback.SetSensitive(FALSE);

	if (objects[target].hasmenu)
		readMenu(); 
	else
		defaultMenuWindow();

	xvisMenus.Open();
}

// ---------------------------------------------
// close and clear up

void closeMenuWindow(int mode)
{
	int ref=0, j=0;
	char tempbuffer[40];

	xvisMenus.Close();

	if (mode==XVMENUOK) {
		if (objects[target].hasmenu) {
			objects[target].menubar->Hide();
			for (int i=0; i<objects[target].nummenu; i++) 
				delete objects[target].pulldown[i];
			for (int i=0; i<objects[target].totalmenu; i++)
				if (objects[target].menuinfo[i].codebuffer != NULL) {
					delete objects[target].menuinfo[i].codebuffer;
					objects[target].menuinfo[i].codebuffer=NULL;
				}
			delete objects[target].menubar;
			delete objects[target].menuinfo;
			objects[target].hasmenu=FALSE;
			objects[target].nummenu=0;	
			objects[target].totalmenu=0;
		}
		if ((!objects[target].hasmenu) && (numeditlist>1) && (nummain>0)) {
			objects[target].totalmenu=numeditlist-1;
			objects[target].nummenu=nummain;
			for (int i=0; i<nummain; i++)
				objects[target].numpulldown[i]=numsub[i];

			objects[target].menuinfo=new struct menuentry[numeditlist-1];
			for (int i=0; i<numeditlist-1; i++) {
				ref=menuchecklist[i];
				objects[target].menuinfo[i].type=editlist[ref].type;
				objects[target].menuinfo[i].listpoint=editlist[ref].listpoint;
				strcpy(objects[target].menuinfo[i].name, editlist[ref].name);
				strcpy(objects[target].menuinfo[i].callback, editlist[ref].callback);
				strcpy(tempbuffer, editlist[ref].caption);
				if (editlist[ref].type==XVMENUSUB) {
					for (j=5; j<strlen(editlist[ref].caption); j++)
						tempbuffer[j-5]=editlist[ref].caption[j];
					tempbuffer[j-5]='\0';
				}
				strcpy(objects[target].menuinfo[i].caption, tempbuffer);
				objects[target].menuinfo[i].codebuffer=NULL;
				if (editlist[ref].codebuffer!=NULL) {
					objects[target].menuinfo[i].codebuffer=
						new char[strlen(editlist[ref].codebuffer)+10];
					strcpy(objects[target].menuinfo[i].codebuffer,
						editlist[ref].codebuffer);
					delete editlist[ref].codebuffer;
					editlist[ref].codebuffer=NULL;
				}
			}
			constructVisualMenu(target, numeditlist-1);
			objects[target].hasmenu=TRUE;	
		}
	}
	delete editlist;
	delete menueditlist;
}

// --------------------------------------------
// construct the menu on screen 

void constructVisualMenu(int targ, int numberof)
{
	int build_count=0, etype=0;
	Pixel temppix;

	objects[targ].changed=TRUE;
	for (int i=0; i<numberof; i++) 
		if (objects[targ].menuinfo[i].type==XVMENUMAIN)
			menu_build[build_count++]=objects[targ].menuinfo[i].caption;

	makeMenuBar(targ, build_count, menu_build);			
			
	build_count=0;
	Boolean flag=FALSE;
	int vcount=0;
	int mcount=0;
	char bname[40];

	strcpy(bname, objects[targ].menuinfo[0].caption);
	while (build_count<=numberof) {
		etype=objects[targ].menuinfo[build_count].type;
		if ((build_count==numberof) || (etype==XVMENUMAIN)) 
			if (flag) {
				objects[targ].pulldown[mcount]= new xvPulldownMenu(
					bname, vcount, menu_build, dummy_list, 
					objects[targ].menubar->GetButton(bname));
 
				objects[targ].pulldown[mcount]->ChangeFont(
					objects[targ].obj_menu_font);
				vcount=0;
				mcount++;
				strcpy(bname, objects[targ].menuinfo[build_count].caption);
					flag=FALSE;
			}
		if ((etype==XVMENUSUB) || (etype==XVMENUSEP)) {
			flag=TRUE;
			menu_build[vcount]=objects[targ].menuinfo[build_count].caption;
			vcount++;
		}
		build_count++;
	}
}

// -----------------------------------------------
// Make a menubar...

void makeMenuBar(int targ, int num, char **args)
{
	Pixel temppix;

	objects[targ].menubar=new xvMenuBar("mbar", *objects[targ].pwindow, num, args, FALSE);
	objects[targ].menubar->SetDistance(0,0);
	objects[targ].menubar->SetDefaultDistance(0);
	objects[targ].menubar->SetFixedPosition(TRUE);
	objects[targ].menubar->SetWidth(objects[targ].pcore->GetWidth()-7);
	objects[targ].menubar->SetHeight(35);
	objects[targ].menubar->SetAllResource(XtNshadowWidth, 0);
	objects[targ].menubar->ChangeFont(objects[targ].obj_menu_font);
	objects[targ].menubar->SetAllResource(XtNresizable, TRUE);
	if (strcmp(objects[targ].menu_background, "(default)")!=0) {
		temppix=stringToPixel(*objects[targ].menubar, objects[targ].menu_background);
		if (temppix != -1) {
			objects[targ].menubar->SetBackground(temppix);
			objects[targ].menubar->SetAllResource(XtNbackground, (XtArgVal) temppix);
		}
	}
	if (strcmp(objects[targ].menu_foreground, "(default)")!=0) {
		temppix=stringToPixel(*objects[targ].menubar, objects[targ].menu_foreground);
		if (temppix != -1) {
			objects[targ].menubar->SetForeground(temppix);
			objects[targ].menubar->SetAllResource(XtNforeground, (XtArgVal) temppix);
		}
	}
	objects[targ].menubar->ManageChildren();
	objects[targ].pcore->SetResizable(TRUE);	
}

// -----------------------------------------------
// find list position of (pos)

int findListPosition(int pos)
{
	int location=-1;
	for (int i=0; i<MAXMENU; i++)
		if ((editlist[i].type==XVMENUMAIN) || (editlist[i].type==XVMENUSUB) || (editlist[i].type==XVMENUSEP))
			if (editlist[i].listpoint==pos) {
				location=i;
				break;
			}
	return(location);
}

// -----------------------------------------------
// find void entry

int findVoidEntry()
{
	int location=-1;

	for (int i=0; i<MAXMENU; i++)
		if (editlist[i].type==XVMENUVOID) {
			location=i;
			break;
		}

	return(location);
}

// -----------------------------------------------
// find specific entry

int findEntry(String caption, String name, String callback)
{
	Boolean found=FALSE;
	Boolean find_caption=FALSE, find_name=FALSE, find_call=FALSE;
	int location=-1;

	if (caption!=NULL) find_caption=TRUE;
	if (name!=NULL) find_name=TRUE;
	if (callback!=NULL) find_call=TRUE;

	for (int i=0; i<MAXMENU; i++) 
		if (editlist[i].type!=XVMENUVOID) {
			if (find_caption) {
				if (strcmp(caption, editlist[i].caption)==0) 
					found=TRUE;
				else
					found=FALSE;
			}
			if (find_name) {
				if (strcmp(name, editlist[i].name)==0)
					found=TRUE;
				else
					found=FALSE;
			}
			if (find_call) { 
				if (strcmp(callback, editlist[i].callback)==0)
					found=TRUE;
				else
					found=FALSE;
			}

			if (found) {
				location=i;
				break;
			}
		}
	return(location);
}

//---------------------------------------------
// update entry
 
void updateListEntry()
{
	updateListInfo();
	int loc=selected_entry;
	if (loc!=-1) 
		if ((editlist[loc].type!=XVMENUEND) && (editlist[loc].type!=XVMENUSEP)) {
			strcpy(editlist[loc].name, menu_name);
			strcpy(editlist[loc].caption, menu_caption);
			strcpy(editlist[loc].callback, menu_callback);
			buildList();
			return;
		}

	dataAlert("Update", "Cannot update this  menu object!");		
}

// ---------------------------------------------
// insert entry

void insertListEntry(int etype)
{
	updateListInfo();
	int loc=findVoidEntry();
	if (loc!=-1) {
		editlist[loc].type=etype;
		if (selected_entry==-1) {
			editlist[loc].listpoint=numeditlist-1;
			selected_entry=numeditlist-1;
		}
		else
			editlist[loc].listpoint=selected_entry;

		strcpy(editlist[loc].name, menu_name);
		strcpy(editlist[loc].caption, menu_caption);
		strcpy(editlist[loc].callback, menu_callback);
		for (int i=0; i<MAXMENU; i++) 
			if ((i!=loc) && (editlist[i].type!=XVMENUVOID) && 
				(editlist[i].listpoint>=selected_entry)) editlist[i].listpoint++;
		buildList();
	}
	else
		dataAlert("Add", "Cannot add another menu object!\nOut Of Memory!");		

}

// ----------------------------------------------

int parseMenuData()
{
	int ok=XVMENUNOERR;
	Boolean in_sub=FALSE;
	int ptr, lasttype=XVMENUVOID;
	
	nummain=0;

	for (int i=0; i<numeditlist-1; i++) {
		ptr=menuchecklist[i];
		if (! in_sub)
			if (editlist[ptr].type==XVMENUMAIN) {
				in_sub=TRUE;
				nummain++;
				numsub[nummain-1]=0;
				lasttype=XVMENUMAIN;
			}
			else { 
				if (lasttype==XVMENUVOID) { ok=XVMENUNOMAIN; break; }
				if (editlist[ptr].type==XVMENUSEP) { ok=XVMENUNOMAIN; break; };
				in_sub=TRUE;
				nummain++;
				numsub[nummain-1]=1;
				lasttype=XVMENUSUB;
			}
		else 
			if (editlist[ptr].type==XVMENUMAIN) {
				if (numsub[nummain-1]==0) {
					ok=XVMENUNOSUB;
					break;
				}
				in_sub=FALSE;
				lasttype=XVMENUMAIN;
			}
			else {
				numsub[nummain-1]++;
				lasttype=XVMENUSUB;
			}
	}
	if ((in_sub) && (numsub[nummain-1]==0)) ok=XVMENUNOSUB;
	if ((!in_sub) && (lasttype=XVMENUMAIN)) ok=XVMENUNOSUB;
	if ((ok==XVMENUNOERR) && (nummain==-1)) ok=XVMENUERROR;
	return(ok);
}

// ----------------------------------------------
		